/// implement TaskDialog window (native on Vista/Seven, emulated on XP)
// - this unit is a part of the freeware Synopse mORMot framework,
// licensed under a MPL/GPL/LGPL tri-license; version 1.18
unit SynTaskDialog;

{
    This file is part of Synopse framework.

    Synopse framework. Copyright (C) 2013 Arnaud Bouchez
      Synopse Informatique - http://synopse.info

  *** BEGIN LICENSE BLOCK *****
  Version: MPL 1.1/GPL 2.0/LGPL 2.1

  The contents of this file are subject to the Mozilla Public License Version
  1.1 (the "License"); you may not use this file except in compliance with
  the License. You may obtain a Copy of the License at
  http://www.mozilla.org/MPL

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the License.

  The Original Code is Synopse framework.

  The Initial Developer of the Original Code is Arnaud Bouchez.

  Portions created by the Initial Developer are Copyright (C) 2013
  the Initial Developer. All Rights Reserved.

  Contributor(s):
  Alternatively, the contents of this file may be used under the terms of
  either the GNU General Public License Version 2 or later (the "GPL"), or
  the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
  in which case the provisions of the GPL or the LGPL are applicable instead
  of those above. If you wish to allow use of your version of this file only
  under the terms of either the GPL or the LGPL, and not to allow others to
  use your version of this file under the terms of the MPL, indicate your
  decision by deleting the provisions above and replace them with the notice
  and other provisions required by the GPL or the LGPL. If you do not delete
  the provisions above, a recipient may use your version of this file under
  the terms of any one of the MPL, the GPL or the LGPL.

  ***** END LICENSE BLOCK *****

  Version 1.13
  - initial release

  Version 1.15
  - new tdfQueryMasked function to display * in the tdfQuery editor field

  Version 1.16
  - fixed issue when changing the current application with Alt+Tab - see
    http://synopse.info/fossil/tktview?name=01395e5932
  - fixed compiler error when using the unit with runtime packages enabled
    (known compiler issue about string resources, referenced as E2201)
  - default modal dialog parent changed into any current active form
  - added tdfQueryFieldFocused optional flag to focus the input field component
  - some aesthetical rendering changes and code clean-up (e.g. no temporary
    form necessary), thanks to uligerhardt proposals

  Version 1.18
  - fixed label height display when long text is wrapped on several lines
  - bottom buttons use better looking TButton component
  - bottom buttons won't Trim expected shortcut definition, in emulated mode
  - now compiles and run in Win64 platform (Delphi XE2+)

}

interface

{$IFDEF CONDITIONALEXPRESSIONS}  // Delphi 6 or newer
  {$ifndef VER140} // Delphi 6
    {$define WITHUXTHEME} // Themes unit exists till Delphi 7
  {$endif}
{$ENDIF}

uses
  Windows, Classes, SysUtils, Messages, UxTheme, Math, CommCtrl,
  {$IF CompilerVersion >= 20.0}
  ShellAPI,
  {$IFEND}
  {$IF CompilerVersion >= 23.0}
  Vcl.Consts, Vcl.ClipBrd, Vcl.Themes, Vcl.Dialogs, Vcl.Menus, Vcl.ComCtrls,
  Vcl.Graphics, Vcl.Forms, Vcl.Controls, Vcl.StdCtrls, Vcl.ExtCtrls, Vcl.Buttons,
  {$ELSE}
  Consts, ClipBrd, Themes, Dialogs, Menus, ComCtrls,
  Graphics, Forms, Controls, StdCtrls, ExtCtrls, Buttons,
  {$IFEND}
  {$ifdef USETMSPACK}
  AdvGlowButton, AdvMenus, TaskDialog, TaskDialogEx,
  {$endif}
  uExpandButton, uCommandLink, uTaskProgress, uWindowsVersion;

var
  /// will map a generic OK picture from SynTaskDialog.res
  BitmapOK: TBitmap;
  /// will map a generic Arrow picture from SynTaskDialog.res
  BitmapArrow: TBitmap;

  /// will map a default font, according to the available
  // - if Calibri is installed, will use it
  // - will fall back to Tahoma otherwise
  DefaultFont: TFont;

{$ifndef USETMSPACK}
  /// is filled once in the initialization block below
  // - you can set this reference to nil to force Delphi dialogs even
  // on Vista/Seven (e.g. make sense if TaskDialogBiggerButtons=True)
  TaskDialogIndirect: function(AConfig: Pointer; Res: PInteger;
    ResRadio: PInteger; VerifyFlag: PBOOL): HRESULT; stdcall;

  {$IFNDEF UNICODE}
  GetThemeFont: function(hTheme: HTHEME; hdc: HDC; iPartId, iStateId, iPropId: Integer;
    var pFont: LOGFONTW): HRESULT; stdcall;
  {$ENDIF}

type
  /// the standard kind of common buttons handled by the Task Dialog
  TCommonButton = (
    cbOK, cbYes, cbNo, cbRetry, cbCancel, cbClose);

  /// set of standard kind of common buttons handled by the Task Dialog
  TCommonButtons = set of TCommonButton;

  /// the available main icons for the Task Dialog
  TTaskDialogIcon = (
    tiBlank, tiWarning, tiQuestion, tiError, tiInformation, tiNotUsed, tiShield);

  /// the available footer icons for the Task Dialog
  TTaskDialogFooterIcon = (
    tfiBlank, tfiWarning, tfiQuestion, tfiError, tfiInformation, tfiNotUsed, tfiShield);

  /// the available configuration flags for the Task Dialog
  // - most are standard TDF_* flags used for Vista/Seven native API
  // (see http://msdn.microsoft.com/en-us/library/bb787473(v=vs.85).aspx
  // for TASKDIALOG_FLAGS)
  // - tdfQuery and tdfQueryMasked are custom flags, implemented in pure Delphi
  // code to handle input query
  // - our emulation code will handle only tdfUseCommandLinks,
  // tdfUseCommandLinksNoIcon, and tdfQuery options
  TTaskDialogFlag = (
    tdfEnableHyperLinks, tdfUseHIconMain, tdfUseHIconFooter,
    tdfAllowDialogCancellation, tdfUseCommandLinks, tdfUseCommandLinksNoIcon,
    tdfExpandFooterArea, tdfExpandByDefault, tdfVerificationFlagChecked,
    tdfShowProgressBar, tdfShowMarqueeProgressBar, tdfCallbackTimer,
    tdfPositionRelativeToWindow, tdfRtlLayout, tdfNoDefaultRadioButton,
    tdfCanBeMinimized, tdfQuery, tdfQueryMasked, tdfQueryFieldFocused);

  /// set of available configuration flags for the Task Dialog
  TTaskDialogFlags = set of TTaskDialogFlag;

  /// implements a TaskDialog
  // - will use the new TaskDialog API under Vista/Seven, and emulate it with
  // pure Delphi code and standard themed VCL components under XP or 2K
  // - create a TTaskDialog object/record on the stack will initialize all
  // its string parameters to '' (it's a SHAME that since Delphi 2009, objects
  // are not initialized any more: we have to define this type as object before
  // Delphi 2009, and as record starting with Delphi 2009)
  // - set the appropriate string parameters, then call Execute() with all
  // additional parameters
  // - RadioRes/SelectionRes/VerifyChecked will be used to reflect the state
  // after dialog execution
  // - here is a typical usage:
  // !var Task: TTaskDialog;
  // !begin
  // !  Task.Inst := 'Saving application settings';
  // !  Task.Content := 'This is the content';
  // !  Task.Radios := 'Store settings in registry'#10'Store settings in XML file';
  // !  Task.Verify := 'Do no ask for this setting next time';
  // !  Task.VerifyChecked := True;
  // !  Task.Footer := 'XML file is perhaps a better choice';
  // !  Task.Execute([],0,[],tiQuestion,tfiInformation,200);
  // !  ShowMessage(IntToStr(Task.RadioRes)); // 200=Registry, 201=XML
  // !  if Task.VerifyChecked then
  // !    ShowMessage(Task.Verify);
  // !end;
  TTaskDialog = {$ifdef UNICODE}record{$else}object{$endif}
    /// the main title of the dialog window
    // - if left void, the title of the application main form is used
    Title: string;
    /// the main instruction (first line on top of window)
    // - any '\n' will be converted into a line feed
    // - if left void, the text is taken from the current dialog icon kind
    Inst: string;
    /// the dialog's primary content content text
    // - any '\n' will be converted into a line feed
    Content: string;
    /// a #13#10 or #10 separated list of custom buttons
    // - they will be identified with an ID number starting at 100
    // - by default, the buttons will be created at the dialog bottom, just
    // like the common buttons
    // - if tdfUseCommandLinks flag is set, the custom buttons will be created
    // as big button in the middle of the dialog window; in this case, any
    // '\n' will be converted as note text (shown with smaller text under native
    // Vista/Seven TaskDialog, or as popup hint within Delphi emulation)
    Buttons: string;
    /// a #13#10 or #10 separated list of custom radio buttons
    // - they will be identified with an ID number starting at 200
    // - aRadioDef parameter can be set to define the default selected value
    // - '\n' will be converted as note text (shown with smaller text under
    // native Vista/Seven TaskDialog, or as popup hint within Delphi emulation)
    Radios: string;
    /// the expanded information content text
    // - any '\n' will be converted into a line feed
    // - the Delphi emulation will always show the Info content (there is no
    // collapse/expand button)
    Info: string;
    /// the button caption to be displayed when the information is collapsed
    // - not used under XP: the Delphi emulation will always show the Info content
    InfoExpanded: string;
    /// the button caption to be displayed when the information is expanded
    // - not used under XP: the Delphi emulation will always show the Info content
    InfoCollapse: string;
    /// the footer content text
    // - any '\n' will be converted into a line feed
    Footer: string;
    /// the text of the bottom most optional checkbox
    Verify: string;
    /// a #13#10 or #10 separated list of items to be selected
    // - if set, a Combo Box will be displayed to select
    // - if tdfQuery is in the flags, the combo box will be in edition mode,
    // and the user will be able to edit the Query text or fill the field
    // with one item of the selection
    // - this selection is not handled via the Vista/Seven TaskDialog, but
    // with our Delphi emulation code (via a TComboBox)
    Selection: string;
    /// some text to be edited
    // - if tdfQuery is in the flags, will contain the default query text
    // - if Selection is set, the
    Query: string;
    /// the selected radio item
    // - first is numeroted 0
    RadioRes: Integer;
    /// after execution, contains the selected item from the Selection list
    SelectionRes: Integer;
    /// reflect the the bottom most optional checkbox state
    // - if Verify is not '', should be set before execution
    // - after execution, will contain the final checkbox state
    VerifyChecked: BOOL;
    MainIconHandle: HICON;
    FooterIconHandle: HICON;
    /// launch the TaskDialog form
    // - some common buttons can be set via aCommonButtons
    // - in emulation mode, aFlags will handle only tdfUseCommandLinks,
    // tdfUseCommandLinksNoIcon, and tdfQuery options
    // - will return 0 on error, or the Button ID (e.g. mrOk for the OK button
    // or 100 for the first custom button defined in Buttons string)
    // - if Buttons was defined, aButtonDef can set the selected Button ID
    // - if Radios was defined, aRadioDef can set the selected Radio ID
    // - aDialogIcon and aFooterIcon are used to specify the displayed icons
    // - aWidth can be used to force a custom form width (in pixels)
    // - aParent can be set to any HWND - by default, Application.DialogHandle
    // - if aNonNative is TRUE, the Delphi emulation code will always be used
    // - aEmulateClassicStyle can be set to enforce conformity with the non themed
    // user interface - see @http://synopse.info/forum/viewtopic.php?pid=2867#p2867
    function Execute(aCommonButtons: TCommonButtons = [];
      aButtonDef: Integer = 0; aFlags: TTaskDialogFlags = [];
      aDialogIcon: TTaskDialogIcon = tiInformation;
      aFooterIcon: TTaskDialogFooterIcon = tfiWarning;
      aRadioDef: Integer = 0; aWidth: Integer = 0; aParent: HWND = 0;
      aNonNative: boolean = False; aEmulateClassicStyle: boolean = False): Integer;
    private
      TaskDialogHandle: HWND;
      AltDialog: TForm;
      FMax: Integer;
      FState: TProgressBarState;
      FMin: Integer;
      FMarqueeSpeed: Cardinal;
      FPosition: Integer;
      FProgress: TTaskDialogProgress;
      procedure SetMarqueeSpeed(const Value: Cardinal);
      procedure SetMax(const Value: Integer);
      procedure SetMin(const Value: Integer);
      procedure SetPosition(const Value: Integer);
      procedure SetState(const Value: TProgressBarState);
    public
      procedure InitProgress;
      procedure EndDialog(nResult: Integer = mrCancel);
      property MarqueeSpeed: Cardinal read FMarqueeSpeed write SetMarqueeSpeed;
      property ProgressMax: Integer read FMax write SetMax;
      property ProgressMin: Integer read FMin write SetMin;
      property ProgressPosition: Integer read FPosition write SetPosition;
      property ProgressState: TProgressBarState read FState write SetState;
  end;
  PTaskDialog = ^TTaskDialog;

  /// a wrapper around the TTaskDialog.Execute method
  // - used to provide a "flat" access to task dialog parameters
  TTaskDialogEx = {$ifdef UNICODE}record{$else}object{$endif}
    /// the associated main TTaskDialog instance
    Base: TTaskDialog;
    /// some common buttons to be displayed
    CommonButtons: TCommonButtons;
    /// the default button ID
    ButtonDef: Integer;
    /// the associated configuration flags for this Task Dialog
    // - in emulation mode, aFlags will handle only tdfUseCommandLinks,
    // tdfUseCommandLinksNoIcon, and tdfQuery options
    Flags: TTaskDialogFlags;
    /// used to specify the dialog icon
    DialogIcon: TTaskDialogIcon;
    /// used to specify the footer icon
    FooterIcon: TTaskDialogFooterIcon;
    /// the default radio button ID
    RadioDef: Integer;
    /// can be used to force a custom form width (in pixels)
    Width: Integer;
    /// if TRUE, the Delphi emulation code will always be used
    NonNative: boolean;
    /// can be used to enforce conformity with the non themed user interface
    EmulateClassicStyle: boolean;
    /// main (and unique) method showing the dialog itself
    // - is in fact a wrapper around the TTaskDialog.Execute method
    function Execute(aParent: HWND = 0): Integer;
    procedure Init;
  end;
{$endif USETMSPACK}

type
{$ifdef USETMSPACK}
  /// a TMS PopupMenu
  TSynPopupMenu = TAdvPopupMenu;

  TSynButtonParent = TAdvGlowButton;
  TSynTaskDialogButton = TAdvGlowButton;
{$else}
  /// a generic VCL popup menu
  TSynPopupMenu = TPopupMenu;

//TSynButtonParent = {$ifdef WITHUXTHEME}TBitBtn{$else}TButton{$endif};
  TSynButtonParent = TCommandLink;
  TSynTaskDialogButton = TButton;
{$endif USETMSPACK}

  /// a generic Button to be used in the User Interface
  // - is always a Themed button: under Delphi 6, since TBitBtn is not themed,
  // it will be a row TButton with no glyph... never mind...
  TSynButton = class(TSynButtonParent)
  protected
    {$ifndef USETMSPACK}
    fDropDownMenu: TSynPopupMenu;
    {$endif}
  public
    /// create a standard button instance
    // - ModalResult/Default/Cancel properties will be set as exepcted for this
    // kind of button
    constructor CreateKind(Owner: TWinControl; Btn: TCommonButton;
      Left, Right, Width, Height: Integer);
    /// set the glyph of the button
    // - set nothing under Delphi 6
    procedure SetBitmap(Bmp: TBitmap);
{$ifndef USETMSPACK}
    /// drop down the associated Popup Menu
    procedure DoDropDown;
    /// the associated Popup Menu to drop down
    property DropDownMenu: TSynPopupMenu read fDropDownMenu write fDropDownMenu;
{$endif}
  end;

/// return the text without the '&' characters within
function UnAmp(const s: string): string;

{$ifndef USETMSPACK}
var
{
  /// if set to TRUE, buttons will be bigger than default
  // - can be useful e.g. for touch screens
  // - will work only for the Delphi emulated version (aNonNative=True) of
  // TSynTask - could be combined with @TaskDialogIndirect := nil;
  TaskDialogBiggerButtons: boolean = False;
}
  /// a default Task Dialog wrapper instance
  // - can be used to display some information with less parameters
  DefaultTaskDialog: TTaskDialogEx = (
    (*
    CommonButtons: [];
    ButtonDef: 0;
    Flags: [];
    *)
    DialogIcon: tiInformation;
    FooterIcon: tfiWarning;
    (*
    RadioDef: 0;
    Width: 0;
    NonNative: False;
    EmulateClassicStyle: False;
    *)
    );
{$endif}

implementation

{$R SynTaskDialog.res}

const
  TD_BTNMOD: array[TCommonButton] of Integer = (
    mrOk, mrYes, mrNo, mrRetry, mrCancel, mrAbort);

function TD_BTNS(button: TCommonButton): Pointer;
begin
  case button of
    cbOK:     result := @SMsgDlgOK;
    cbYes:    result := @SMsgDlgYes;
    cbNo:     result := @SMsgDlgNo;
    cbCancel: result := @SMsgDlgCancel;
    cbRetry:  result := @SMsgDlgRetry;
    cbClose:  result := @SCloseButton;
  else
    result := nil;
  end;
end;

{ TSynButton }

constructor TSynButton.CreateKind(Owner: TWinControl; Btn: TCommonButton;
  Left, Right, Width, Height: Integer);
begin
  Create(Owner);
  Parent := Owner;
  SetBounds(Left, Right, Width, Height);
  Caption := LoadResString(TD_BTNS(Btn));
  ModalResult := TD_BTNMOD[Btn];
  case Btn of
    cbOK:     Default := True;
    cbCancel: Cancel := True;
  end;
  case Btn of
    cbOK: SetBitmap(BitmapOK);
  end;
end;

{$ifndef USETMSPACK}
procedure TSynButton.DoDropDown;
begin
  if (DropDownMenu <> nil) then
    with ClientToScreen(BoundsRect.TopLeft) do
      DropDownMenu.Popup(X, Y + Height);
end;
{$endif}

procedure TSynButton.SetBitmap(Bmp: TBitmap);
begin
  if (Bmp <> nil) then
    {$ifdef USETMSPACK}
    Picture.Assign(Bmp);
    {$else}
      {$ifdef WITHUXTHEME}
      Glyph := Bmp;
      {$else}
      // Delphi 6 TBitBtn has no theming -> use generic TButton without glyph
      {$endif}
    {$endif}
end;

function UnAmp(const s: string): string;
begin
  result := StripHotkey(s);
end;

{$ifndef USETMSPACK}

const
  TD_ICONS: array[TTaskDialogIcon] of Integer = (
    17, 84, 99, 98, 81, 0, 78);
  TD_FOOTERICONS: array[TTaskDialogFooterIcon] of Integer = (
    17, 84, 99, 98, 65533, 0, 65532);
  WIN_ICONS: array[TTaskDialogIcon] of PChar = (
    nil, IDI_WARNING, IDI_QUESTION, IDI_ERROR, IDI_INFORMATION, nil, IDI_WINLOGO);
  WIN_FOOTERICONS: array[TTaskDialogFooterIcon] of PChar = (
    nil, IDI_WARNING, IDI_QUESTION, IDI_ERROR, IDI_INFORMATION, nil, IDI_WINLOGO);
  WIN_FOOTERICONS_SMALL: array[TTaskDialogFooterIcon] of Integer = (
    0, 101, 102, 103, 104, 0, 106);

function IconMessage(Icon: TTaskDialogIcon): string;
begin
  case Icon of
    tiWarning:
      result := SMsgDlgWarning;
    tiQuestion:
      result := SMsgDlgConfirm;
    tiError:
      result := SMsgDlgError;
    tiInformation, tiShield:
      result := SMsgDlgInformation;
  else
    result := '';
  end;
end;

procedure InitComCtl6;
begin
  if CheckWin32Version(6) then
    @TaskDialogIndirect := GetProcAddress(GetModuleHandle(comctl32), 'TaskDialogIndirect')
  else
    @TaskDialogIndirect := nil;
end;

{ TTaskDialog }

type
  // see http://msdn.microsoft.com/en-us/library/bb787473
  PTASKDIALOG_BUTTON = ^TTASKDIALOG_BUTTON;
  TTASKDIALOG_BUTTON =
  packed record
    nButtonID: Integer;
    pszButtonText: PWideChar;
  end;

  TTASKDIALOGCONFIG =
  packed record
    cbSize: Integer;
    hwndParent: HWND;
    hInstance: THandle;
    dwFlags: cardinal;
    dwCommonButtons: cardinal;
    pszWindowTitle: PWideChar;
    hMainIcon: HICON;
    pszMainInstruction: PWideChar;
    pszContent: PWideChar;
    cButtons: Integer;
    pButtons: PTASKDIALOG_BUTTON;
    nDefaultButton: Integer;
    cRadioButtons: Integer;
    pRadioButtons: PTASKDIALOG_BUTTON;
    nDefaultRadioButton: Integer;
    pszVerificationText: PWideChar;
    pszExpandedInformation: PWideChar;
    pszExpandedControlText: PWideChar;
    pszCollapsedControlText: PWideChar;
    hFooterIcon: HICON;
    pszFooter: PWideChar;
//  pfCallback: Pointer;
//  lpCallbackData: Pointer;
    pfCallback: TFTaskDialogCallback;
    lpCallbackData: LONG_PTR;
    cxWidth: Integer;
  end;

{ TExpandButtonEventHandlerClass }

type
  TExpandButtonEventHandlerClass = class(TPersistent)
  private
    FExpandButton: TExpandButton;
    FExpandPanel: TPanel;
    FTopPos: Integer;
    FCollapseHeight: Integer;
    FExpandHeight: Integer;
    FExpandLabel: TLabel;
    {$IF CompilerVersion >= 20.0}
    FExpandLinkLabel: TLinkLabel;
    {$IFEND}
  public
    procedure Click(Sender: TObject);
    property ExpandButton: TExpandButton read FExpandButton write FExpandButton;
    property ExpandPanel: TPanel read FExpandPanel write FExpandPanel;
    property ExpandLabel: TLabel read FExpandLabel write FExpandLabel;
    {$IF CompilerVersion >= 20.0}
    property ExpandLinkLabel: TLinkLabel read FExpandLinkLabel write FExpandLinkLabel;
    {$IFEND}
    property TopPos: Integer read FTopPos write FTopPos;
    property CollapseHeight: Integer read FCollapseHeight write FCollapseHeight;
    property ExpandHeight: Integer read FExpandHeight write FExpandHeight;
  end;

procedure TExpandButtonEventHandlerClass.Click(Sender: TObject);
begin
  if ExpandButton.Expand then
    begin
      if (ExpandLabel <> nil) then
        ExpandLabel.Visible := True
      else
        begin
          {$IF CompilerVersion >= 20.0}
           if (ExpandLinkLabel <> nil) then
             ExpandLinkLabel.Visible := True
          {$IFEND}
        end;
      ExpandPanel.Align := alNone;
      ExpandPanel.Top := FTopPos;
      ExpandPanel.Align := alTop;
      ExpandPanel.Height := Self.ExpandHeight;
      ExpandPanel.Parent.Height := ExpandPanel.Parent.Height + (Self.ExpandHeight - Self.CollapseHeight);
    end
  else
    begin
      if (ExpandLabel <> nil) then
        ExpandLabel.Visible := False
      else
        begin
          {$IF CompilerVersion >= 20.0}
           if (ExpandLinkLabel <> nil) then
             ExpandLinkLabel.Visible := False
          {$IFEND}
        end;
      ExpandPanel.Height := Self.CollapseHeight;
      ExpandPanel.Parent.Height := ExpandPanel.Parent.Height - (Self.ExpandHeight - Self.CollapseHeight);
    end;
end;

{ TLinkLabelEventHandlerClass }
{$IF CompilerVersion >= 20.0}
type
  TLinkLabelEventHandlerClass = class(TPersistent)
  public
    procedure LinkClick(Sender: TObject; const Link: string;
      LinkType: TSysLinkType);
  end;

procedure TLinkLabelEventHandlerClass.LinkClick(Sender: TObject; const Link: string;
  LinkType: TSysLinkType);
begin
  ShellExecute(0, 'open', PChar(Link), '', '', SW_SHOWNORMAL);
end;
{$IFEND}


{ TFormEventHandlerClass }
type
  TFormEventHandlerClass = class(TPersistent)
  private
    Edit: TEdit;
    FText: string;
    FExpandButton: TExpandButton;
    FCheckBox: TExpandButton;
  public
    constructor Create;
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    property Text: string read FText write FText;
    property ExpandButton: TExpandButton read FExpandButton write FExpandButton;
    property CheckBox: TExpandButton read FCheckBox write FCheckBox;
  end;

constructor TFormEventHandlerClass.Create;
begin
  Edit         := nil;
  ExpandButton := nil;
  CheckBox     := nil;
  FText        := '';
end;

procedure TFormEventHandlerClass.FormKeyPress(Sender: TObject; var Key: Char);
const
  CheckSign: array [Boolean] of string = (' ', 'X');
  ExpandSign: array [Boolean] of string = ('V', '^');
var
  Dmy, s: string;
begin
  if (Key = ^C) then
    begin
      if (Edit <> nil) and (Edit.Focused) then
        Exit;
      Dmy := FText;
      if (CheckBox <> nil) then
        Dmy := StringReplace(Dmy, '@VerifyChecked@', CheckSign[ExpandButton.Expand], [rfIgnoreCase]);
      if (ExpandButton <> nil) then
        begin
          Dmy := StringReplace(Dmy, '@ExpandButton@' , ExpandSign[ExpandButton.Expand], [rfIgnoreCase]);
          if ExpandButton.Expand then
            s := ExpandButton.CaptionExpand
          else
            s := ExpandButton.CaptionCollapse;
          Dmy := StringReplace(Dmy, '@ExpandCaption@', s, [rfIgnoreCase]);
        end;
      Clipboard.AsText := Dmy;
    end;
end;

{ TCallbackClass }
type
  TCallbackClass = class
  private
    FTaskDialog: PTaskDialog;
    FTaskProgress: TTaskDialogProgress;
  protected
    function CallbackProc(hwnd: HWND; msg: UINT; wParam: WPARAM;
      lParam: LPARAM; lpRefData: LONG_PTR): HResult; dynamic;
  public
    property TaskDialog: PTaskDialog read FTaskDialog write FTaskDialog;
    property TaskProgress: TTaskDialogProgress read FTaskProgress write FTaskProgress;
  end;

function TCallbackClass.CallbackProc(hwnd: HWND; msg: UINT; wParam: WPARAM;
  lParam: LPARAM; lpRefData: LONG_PTR): HResult;
{$IF CompilerVersion >= 20.0}
var
  URL: string;
{$IFEND}
begin
  TaskDialog.TaskDialogHandle := hwnd;
  if (FTaskProgress <> nil) then
    FTaskProgress.ClientNative := hwnd;
  Result := S_OK;
  {$IF CompilerVersion >= 20.0}
  case msg of
    TDN_HYPERLINK_CLICKED:
      begin
        URL := WideString(LPCWSTR(lParam));
        ShellExecute(0, 'open', PChar(URL), '', '', SW_SHOWNORMAL);
      end;
  end;
  {$IFEND}
end;

function TaskDialogCallbackProc(hwnd: HWND; msg: UINT; wParam: WPARAM;
  lParam: LPARAM; lpRefData: LONG_PTR): HResult; stdcall;
begin
  Result := TCallbackClass(Pointer(lpRefData)).CallbackProc(hwnd, msg, wparam, lparam, 0);
end;

{ TTaskDialog }

procedure TTaskDialog.EndDialog(nResult: Integer);
begin
  if (TaskDialogHandle <> 0) then
    PostMessage(TaskDialogHandle, TDM_CLICK_BUTTON, Windows.WPARAM(nResult), 0)
  else if (AltDialog <> nil) then
    AltDialog.ModalResult := nResult;
end;

function TTaskDialog.Execute(aCommonButtons: TCommonButtons;
  aButtonDef: Integer; aFlags: TTaskDialogFlags;
  aDialogIcon: TTaskDialogIcon; aFooterIcon: TTaskDialogFooterIcon;
  aRadioDef, aWidth: Integer; aParent: HWND; aNonNative: boolean;
  aEmulateClassicStyle: boolean): Integer;
type
  WS = {$ifdef UNICODE}string{$else}WideString{$endif};
  TCalcSize = record
    Width: Integer;
    Height: Integer;
  end;
var
  Config: TTASKDIALOGCONFIG;
  i, X, Y, dX, dY, XB, IconBorder, FontHeight: Integer;
  Par: TWinControl;
  PrevPanel, Panel, pnlSelector, pnlButtons, pnlProgress, pnlExpand, pnlFooter: TPanel;
  HasSelectorPnl, hasProgressPnl, HasButtonPnl, HasExpandPnl, HasFooterPnl: Boolean;
  ShowProgress, IsExpandFooter: Boolean;
  CurrTabOrder: TTabOrder;
  Form: TForm;
  Image: TImage;
  Pic: TIcon;
  Bmp: TBitmap;
  Edit: TEdit;
  Combo: TComboBox;
  List: TStrings;
  B: TCommonButton;
  CommandLink: TSynButton;
  Rad: array of TRadioButton;
  Verif: TExpandButton;
  Expand: TExpandButton;
  Progress: TProgressBarEx;
  dCommonButtons: TCommonButtons;
  ebeh: TExpandButtonEventHandlerClass;
  ExpandLabel: TLabel;
  cc: TCallbackClass;
  {$IF CompilerVersion >= 20.0}
  lleh: TLinkLabelEventHandlerClass;
  ExpandLinkLabel: TLinkLabel;
  {$IFEND}
  fmeh: TFormEventHandlerClass;
  AdjustY, OffsetX, OffsetY: Integer;
  ClipboardText, CTLine, dmyStr: string;
  // Params
  HAS_BUTTONS, HAS_RADIOS, HAS_INFO, HAS_VERIFY, HAS_QUERY, HAS_SELECTION, HAS_FOOTER, HAS_COMMANDLINK: Boolean;
  SmallIconSize: Integer;
  dmyDC: hdc;
  dmyRect: TRect;
  dmyCanvas: TCanvas;
  ExpandLen: array [Boolean] of TCalcSize;
  ExpandLength, OExpandLength: TCalcSize;
  VerifyLength, OVerifyLength: TCalcSize;
  dmyFlgs, ButtonPanelHeight, HeightSum, CommandLinkDefaultheight: Integer;
  ParentHWND: HWND;

  function ParseLinks(Caption: string): string;
  const
    ATag = ['A', 'a'];
  var
    P: PChar;
    InTag, IsAtag, MarkUp: Boolean;
  begin
    P := PChar(Caption);
    result := '';
    InTag  := False;
    IsAtag := False;
    MarkUp := False;
    while (P^ <>  #$00) do
      begin
        case P^ of
          '<':
            begin
              InTag := True;
              Inc(P);
              if( P^ = #$00) then
                Break;
              {$IFDEF UNICODE}
              if CharInSet(P^, ATag) then
              {$ELSE}
              if (P^ in ATag) then
              {$ENDIF}
                IsAtag := True
              {$IFDEF UNICODE}
              else if CharInSet(P^, ['/']) then
              {$ELSE}
              else if (P^ in ['/']) then
              {$ENDIF}
                begin
                  Inc(P);
                  if (P^ = #$00) then
                    Break;
                  {$IFDEF UNICODE}
                  if CharInSet(P^, ATag) then
                  {$ELSE}
                  if (P^ in ATag) then
                  {$ENDIF}
                    begin
                      IsAtag := True;
                      MarkUp := False;
                    end
                  else
                    begin
                      IsAtag := False;
                      result := result + '</' + P^
                    end;
                end
              else
                begin
                  IsAtag := False;
                  result := result + '<'+ P^;
                  InTag := False;
                end;
            end;
          '>':
            begin
              if MarkUp then
                result := result + P^
              else
                begin
                  InTag := False;
                  MarkUp := IsAtag;
                end;
            end;
        else
          if not InTag then
            result := result + P^;
        end;
        Inc(P);
      end;
  end;

  function CR(const aText: string): string;
  begin
    if Pos('\n', aText) = 0 then
      result := aText
    else
      result := StringReplace(aText, '\n', #10, [rfReplaceAll]);
  end;

  function GetNextStringLineToWS(var P: PChar): WS;
  var
    S: PChar;
    {$ifndef UNICODE}tmp: string;{$endif}
  begin
    if (P = nil) then
      result := ''
    else
      begin
        S := P;
        while (S[0] >= ' ') do
          Inc(S);
        {$ifdef UNICODE}
        SetString(result, P, S - P);
        result := CR(result);
        {$else}
        SetString(tmp, P, S - P);
        result := WideString(CR(tmp));
        {$endif}
        while ((S^ <> #0) and (S^ < ' ')) do
          Inc(S); // ignore e.g. #13 or #10
        if (S^ <> #0) then
          P := S
        else
          P := nil;
      end;
  end;

  var
    aHint: string;

  function NoCR(const aText: string): string;
  var
    i: Integer;
  begin
    result := aText;
    aHint := '';
    i := Pos('\n', result);
    if (i > 0) then
      begin
        aHint := CR(Copy(result, i + 2, MaxInt));
        SetLength(result, i - 1);
      end;
  end;

  function N(const aText: string): WS;
  begin
    if (aText = '') then
      result := ''
    else
      result := WS(CR(aText));
  end;

  var
    RU: array of Ws;
    RUCount: Integer;
    But: array of TTASKDIALOG_BUTTON;

  procedure AddRU(Text: string; var n: Integer; firstID: Integer);
  var
    P: PChar;
  begin
    if (Text = '') then
      Exit;
    Text := SysUtils.Trim(Text);
    P := @Text[1]; // '\n' handling in GetNextStringLineToWS(P) will change P^
    while (P <> nil) do
      begin
        if (length(RU) <= RUCount) then
          begin
            SetLength(RU , RUCount + 16);
            SetLength(But, RUCount + 16);
          end;
        RU[RUCount] := GetNextStringLineToWS(P);
        with But[RUCount] do
          begin
            nButtonID := n + firstID;
            pszButtonText := Pointer(RU[RUCount]);
          end;
        Inc(n);
        Inc(RUCount);
      end;
  end;

  function AddLabel(Text: string; IsInstruction: boolean): TLabel;
  var
    R: TRect;
    W: Integer;
    Theme: hTheme;
    DC: hdc;
    ColorRef: TColorRef;
    LOGFONT: {$IFDEF UNICODE}TLOGFONT{$ELSE}TLOGFONTW{$ENDIF};
    PartId: Integer;
  begin
    result := TLabel.Create(Form);
    result.Parent := Par;
    result.WordWrap := True;

    if ThemesEnabled then
      begin
        if VistaOrLater then
          begin
            if IsInstruction then
              PartId := TEXT_INSTRUCTION
            else
              PartId := TEXT_BODYTEXT;
            Theme := OpenThemeData(Application.Handle,'TEXTSTYLE');
            DC := GetDC(0);
            try
              GetThemeFont(Theme, DC, PartId, 0, TMT_FONT, LOGFONT);
              GetThemeColor(Theme, PartId, 0, TMT_TEXTCOLOR, ColorRef);
              result.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(LOGFONT);
              result.Font.Color  := ColorRef and (not clSystemColor);
            finally
              ReleaseDC(0, DC);
              CloseThemeData(Theme);
             end;
          end
        else
          begin
            if IsInstruction then
              result.Font.Color  := $993300
            else
              result.Font.Height := FontHeight;
          end;
      end
    else
      begin
        if IsInstruction then
          result.Font.Style  := [fsBold]
        else
          result.Font.Height := FontHeight;
      end;

    if (tdfEnableHyperLinks in aFlags) {$IF CompilerVersion >= 20.0}and (GetFileVersion(comctl32) < $00060000){$IFEND} then
      Text := ParseLinks(Text);

    Text := CR(Text);
    result.AutoSize := False;
    R.Left   := 0;
    R.Top    := 0;
    W        := aWidth - X - 14;
    R.Right  := W;
    R.Bottom := result.Height;
    R.Bottom := DrawText(result.Canvas.Handle, Pointer(Text), -1, R, DT_WORD_ELLIPSIS or DT_CALCRECT or DT_WORDBREAK);
    result.SetBounds(X, Y, W, R.Bottom);
    result.EllipsisPosition := epWordEllipsis;
    result.ShowAccelChar    := False;
    result.Caption          := Text;
    result.Transparent      := True;
    Inc(Y, R.Bottom + 15)
  end;

  {$IF CompilerVersion >= 20.0}
  function AddLinkLabel(Text: string; IsInstruction: boolean): TLinkLabel;
  var
    R: TRect;
    W: Integer;
    dCanvas: TCanvas;
    Theme: hTheme;
    DC: hdc;
    ColorRef: TColorRef;
    LOGFONT: {$IFNDEF UNICODE}TLOGFONT{$ELSE}TLOGFONTW{$ENDIF};
    PartId: Integer;
  begin
    result := TLinkLabel.Create(Form);
    result.Parent := Par;

    if ThemesEnabled then
      begin
        if VistaOrLater then
          begin
            if IsInstruction then
              PartId := TEXT_INSTRUCTION
            else
              PartId := TEXT_BODYTEXT;
            Theme := OpenThemeData(Application.Handle,'TEXTSTYLE');
            DC := GetDC(0);
            try
              GetThemeFont(Theme, DC, PartId, 0, TMT_FONT, LOGFONT);
              GetThemeColor(Theme, PartId, 0, TMT_TEXTCOLOR, ColorRef);
              result.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(LOGFONT);
              result.Font.Color := ColorRef and (not clSystemColor);
            finally
              ReleaseDC(0, DC);
              CloseThemeData(Theme);
             end;
          end
        else
          begin
            if IsInstruction then
              result.Font.Color := $993300
            else
              result.Font.Height := FontHeight;
          end;
      end
    else
      begin
        if IsInstruction then
          result.Font.Style := [fsBold]
        else
          result.Font.Height := FontHeight;
      end;


    if (tdfEnableHyperLinks in aFlags) {$IF CompilerVersion >= 20.0}and (GetFileVersion(comctl32) < $00060000){$IFEND} then
      Text := ParseLinks(Text);

    Text := StringReplace(CR(Text), '&', '&&', [rfReplaceAll]);
    result.AutoSize := False;
    R.Left := 0;
    R.Top := 0;
    W := aWidth - X - 14;
    R.Right := W;
    R.Bottom := result.Height;
    dCanvas := TCanvas.Create;
    try
      DC := GetDC(result.Handle);
      dCanvas.Handle := DC;
      dCanvas.Font.Assign(result.Font);
      R.Bottom := DrawText(dCanvas.Handle, Pointer(ParseLinks(Text)), -1, R, DT_WORD_ELLIPSIS or DT_CALCRECT or DT_WORDBREAK);
      ReleaseDC(result.Handle, DC);
    finally
      dCanvas.Free;
    end;
    result.SetBounds(X, Y, W, R.Bottom);
    result.Caption := Text;
    result.OnLinkClick := lleh.LinkClick;
    result.ParentColor := True;
    Inc(Y, R.Bottom + 15)
  end;
  {$IFEND}

  function AddButton(s: string; ModalResult: Integer): TSynTaskDialogButton;
  var
    WB, btnWidth: Integer;
  begin
    btnWidth := Form.Canvas.TextWidth(s);
    if ThemesEnabled then
      WB := Math.Max(104, btnwidth + 42)
    else
      WB := Math.Max( 72, btnwidth + 43);
    dec(XB, WB);
    result := TSynTaskDialogButton.Create(Form);
    result.Parent := Par;
    if ThemesEnabled then
      result.SetBounds(XB - 4, 9, WB - 6, 26)
    else
      result.SetBounds(XB, 12, WB - 10, 23);
    result.Caption := s;
    result.ModalResult := ModalResult;
    result.TabOrder := CurrTabOrder;
    case ModalResult of
      mrOk:
        begin
          result.Default := True;
          if (aCommonButtons = [cbOk]) then
            result.Cancel := True;
        end;
      mrCancel:
        result.Cancel := True;
    end;
    if (ModalResult = aButtonDef) then
      Form.ActiveControl := result;
  end;

  function AddLine(AParent: TWinControl; IsSecondaryPanel: Boolean): TImage;
  var
    Theme: hTheme;
  begin
    result := nil;
    if IsSecondaryPanel and (not ThemesEnabled) then
      Exit;
    result := TImage.Create(Form);
    with result do
      begin
        Parent := AParent;
        Align  := alTop;
        Width  := (Parent as Tpanel).Width;
        if (not IsSecondaryPanel) and ThemesEnabled then
          Height := 2
        else
          Height := 1;
        if (ThemesEnabled and VistaOrLater) then
          begin
            Theme := OpenThemeData(Application.Handle,'TASKDIALOG');
            try
              DrawThemeBackground(Theme, result.Canvas.Handle, TDLG_FOOTNOTESEPARATOR, 0, Rect(0, 0, result.Width, result.Height), nil);
            finally
              CloseThemeData(Theme);
            end;
          end
        else
          begin
            result.Canvas.Pen.Style := psSolid;
            result.Canvas.Pen.Color := clBtnShadow;
            result.Canvas.MoveTo(0, 0);
            result.Canvas.LineTo(result.Width, 0);
            if (not IsSecondaryPanel) and ThemesEnabled then
              begin
                result.Canvas.Pen.Color := clBtnHighlight;
                result.Canvas.MoveTo(0, 1);
                result.Canvas.LineTo(result.Width, 1);
              end;
          end;
      end;
  end;

  procedure AddClipboardText(aSection, aText: string; HyperLink: Boolean = False);
  var
    Dmy: String;
  begin
    Dmy := aText;
    if HyperLink then
      Dmy := ParseLinks(Dmy);
    ClipboardText := ClipboardText + aSection + #$0D#$0A + Dmy + #$0D#$0A#$0D#$0A;
  end;

  procedure PreCalcWidth;
  var
    dmyFlgs: Integer;
    dmyRect: TRect;
  begin
    dmyFlgs := DT_VCENTER or DT_NOCLIP or DT_HIDEPREFIX or DT_WORDBREAK or DT_CALCRECT;
    if HAS_INFO then
      begin
        dmyRect := Rect(0, 0, Form.Canvas.TextWidth(InfoExpanded), 24);
        DrawText(Form.Canvas.Handle, Pointer(InfoExpanded), -1, dmyRect, dmyFlgs);
        ExpandLen[False].Width  := dmyRect.Right;
        ExpandLen[False].Height := dmyRect.Bottom;

        dmyRect := Rect(0, 0, Form.Canvas.TextWidth(InfoCollapse), 24);
        DrawText(Form.Canvas.Handle, Pointer(InfoCollapse), -1, dmyRect, dmyFlgs);
        ExpandLen[True].Width  := dmyRect.Right;
        ExpandLen[True].Height := dmyRect.Bottom;

        ExpandLength.Width  := Math.Max(ExpandLen[False].Width , ExpandLen[True].Width );
        ExpandLength.Height := Math.Max(ExpandLen[False].Height, ExpandLen[True].Height);
      end;
    if HAS_VERIFY then
      begin
        dmyRect := Rect(0, 0, Form.Canvas.TextWidth(Verify), 24);
        DrawText(Form.Canvas.Handle, Pointer(Verify), -1, dmyRect, dmyFlgs);
        VerifyLength.Width  := dmyRect.Right;
        VerifyLength.Height := dmyRect.Bottom;
      end;
  end;

  function CalcDialogWidth: Integer;
  var
    i: Integer;
    BtnCaptionWidth: Integer;
    BottomBtnWidth: Integer;
    TotalWidth: Integer;
    cb: TCommonButton;
    s: string;
  begin
    result := 360;
    // Calc CheckBox
    BottomBtnWidth := 0;
    if HAS_INFO then
      BottomBtnWidth := ExpandLength.Width + 23;
    if HAS_VERIFY then
      BottomBtnWidth := Math.Max(BottomBtnWidth, VerifyLength.Width + 23);
    // Calc Common Button
    if (not HAS_BUTTONS) or (HAS_BUTTONS and HAS_COMMANDLINK) then
      begin
        TotalWidth := 0;
        for cb := High(cb) downto Low(cb) do
          if (cb in aCommonButtons) then
            begin
              s := LoadResString(TD_BTNS(cb));
              BtnCaptionWidth := Form.Canvas.TextWidth(s);
              Inc(TotalWidth, Math.Max(104, BtnCaptionWidth + 42));
            end;
        result := Math.Max(result, TotalWidth + BottomBtnWidth + 32);
      end;
    // Calc Custom Button
    if (HAS_BUTTONS and (not HAS_COMMANDLINK)) then
      begin
        TotalWidth := 0;
        with TStringList.Create do
          try
            Text := SysUtils.Trim(Buttons);
            for i:=0 to Count - 1 do
              begin
                s := Trim(Strings[i]);
                BtnCaptionWidth := Form.Canvas.TextWidth(s);
                Inc(TotalWidth, Math.Max(104, BtnCaptionWidth + 42));
              end;
          finally
            Free;
          end;
        result := Math.Max(result, TotalWidth + BottomBtnWidth + 32);
      end;
    // calc Command Link
    if HAS_COMMANDLINK then
      result := Math.Max(result, 396);
  end;

  function GetPanelPosition: Integer;
  begin
    result := PrevPanel.Top + PrevPanel.Height;
  end;

  procedure SetPositionToCenter;
  var
    X, Y: Integer;
    ParentRect: TRect;
  begin
    if Form.Position <> poOwnerFormCenter then
      Exit;
    if (aParent <> 0) and GetWindowRect(aParent, ParentRect) then
      begin
        X := (((ParentRect.Right  - ParentRect.Left) - Form.Width ) div 2) + ParentRect.Left;
        Y := (((ParentRect.Bottom - ParentRect.Top ) - Form.Height) div 2) + ParentRect.Top;
      end
    else
      begin
        X := (Screen.Width  - Form.Width ) div 2;
        Y := (Screen.Height - Form.Height) div 2;
      end;
    if X < Screen.DesktopLeft then
      X := Screen.DesktopLeft;
    if Y < Screen.DesktopTop then
      Y := Screen.DesktopTop;
    Form.Position := poDesigned;
    Form.SetBounds(X, Y, Form.Width, Form.Height);
  end;

begin
  // Params
  HAS_BUTTONS     := (Buttons   <> '');
  if not HAS_BUTTONS then
    aFlags := aFlags - [tdfUseCommandLinks, tdfUseCommandLinksNoIcon];
  HAS_INFO        := (Info      <> '');
  if not HAS_INFO then
    aFlags := aFlags - [tdfExpandFooterArea, tdfExpandByDefault];
  HAS_VERIFY      := (Verify    <> '');
  HAS_QUERY       := (Query     <> '');
  HAS_SELECTION   := (Selection <> '');
  HAS_RADIOS      := (Radios    <> '');
  if not HAS_RADIOS then
    aFlags := aFlags - [tdfNoDefaultRadioButton];
  HAS_FOOTER      := (Footer    <> '');
  if not HAS_FOOTER then
    aFlags := aFlags - [tdfUseHIconFooter];
  HAS_COMMANDLINK := (tdfUseCommandLinks in aFlags) or (tdfUseCommandLinksNoIcon in aFlags);
  TaskDialogHandle := 0;
  AltDialog        := nil;
  FProgress        := nil;
  if (Byte(aCommonButtons) = 0) and (not HAS_BUTTONS) then
    begin
      aCommonButtons := [cbOk];
      if (aButtonDef = 0) then
        aButtonDef := mrOk;
    end;
  if (Title = '') then
    if (Application.MainForm = nil) then
      Title := Application.Title
    else
      Title := Application.MainForm.Caption;
  if (Inst = '') then
    Inst := IconMessage(aDialogIcon);
  if (aParent = 0) then
    if (Screen.ActiveCustomForm <> nil) then
      aParent := Screen.ActiveCustomForm.Handle
    else
      aParent := Application.Handle;

  ShowProgress := (tdfShowProgressBar in aFlags) or (tdfShowMarqueeProgressBar in aFlags);
  if ShowProgress then
    FProgress := TTaskDialogProgress.Create;

  if Assigned(TaskDialogIndirect) and (not aNonNative) and
     (not (tdfQuery in aFlags)) and (not HAS_SELECTION) then
    begin
      // use Vista/Seven TaskDialog implementation (not tdfQuery nor Selection)
      FillChar(Config, SizeOf(Config), 0);
      Config.cbSize                  := SizeOf(Config);
      Config.hwndParent              := aParent;
      Config.pszWindowTitle          := Pointer(N(Title));
      Config.pszMainInstruction      := Pointer(N(Inst));
      Config.pszContent              := Pointer(N(Content));
      RUCount                        := 0;
      AddRU(Buttons, Config.cButtons     , 100);
      AddRU(Radios , Config.cRadioButtons, 200);
      if (Config.cButtons > 0) then
        Config.pButtons := @But[0];
      if (Config.cRadioButtons > 0) then
        Config.pRadioButtons := @But[Config.cButtons];
      Config.pszVerificationText     := Pointer(N(Verify));
      Config.pszExpandedInformation  := Pointer(N(Info));
      Config.pszExpandedControlText  := Pointer(N(InfoExpanded));
      Config.pszCollapsedControlText := Pointer(N(InfoCollapse));
      Config.pszFooter := Pointer(N(Footer));
      if HAS_BUTTONS and (not HAS_COMMANDLINK) then
        aCommonButtons := [];
      dCommonButtons := aCommonButtons;
      dCommonButtons := dCommonButtons - [cbRetry, cbCancel];
      if (cbRetry in aCommonButtons) then
        Include(dCommonButtons, cbCancel);
      if (cbCancel in aCommonButtons) then
        Include(dCommonButtons, cbRetry);
      Config.dwCommonButtons         := Byte(dCommonButtons);
      if HAS_VERIFY and VerifyChecked then
        include(aFlags, tdfVerificationFlagChecked);
      if (Config.cButtons = 0) and (aCommonButtons = [cbOk]) then
        Include(aFlags, tdfAllowDialogCancellation); // just OK -> Esc/Alt+F4 close
      Config.dwFlags                 := Integer(aFlags);
      if (tdfUseHIconMain in aFlags) then
        Config.hMainIcon := MainIconHandle
      else
        Config.hMainIcon := TD_ICONS[aDialogIcon];
      if (tdfUseHIconFooter in aFlags) then
        Config.hFooterIcon := FooterIconHandle
      else
        Config.hFooterIcon := TD_FOOTERICONS[aFooterIcon];
      Config.nDefaultButton          := aButtonDef;
      Config.nDefaultRadioButton     := aRadioDef;
      Config.cxWidth                 := aWidth;

      cc := TCallbackClass.Create;
      try
        cc.TaskDialog := @Self;
        if ShowProgress then
          begin
            FProgress.IsMarquee    := (tdfShowMarqueeProgressBar in aFlags);
            FProgress.Min          := Self.ProgressMin;
            FProgress.Max          := Self.ProgressMax;
            FProgress.Position     := Self.ProgressPosition;
            FProgress.MarqueeSpeed := Self.MarqueeSpeed;
            FProgress.State        := Self.ProgressState;
            cc.TaskProgress        := FProgress;
          end;
        Config.lpCallbackData := LONG_PTR(cc);
        Config.pfCallback     := @TaskDialogCallbackProc;
        if (TaskDialogIndirect(@Config, @result, @RadioRes, @VerifyChecked) = S_OK) then
          Exit; // error (mostly invalid argument) -> execute the VCL emulation
      finally
        TaskDialogHandle := 0;
        if ShowProgress then
          FProgress.Free;
        cc.Free;
      end;
    end;

  // use our native (naive?) Delphi implementation
  Verif := nil;
  Combo := nil;
  Edit  := nil;
  Image := nil;
  Form  := TForm.Create(Application);
  ebeh  := TExpandButtonEventHandlerClass.Create;
  {$IF CompilerVersion >= 20.0}
  lleh  := TLinkLabelEventHandlerClass.Create;
  {$IFEND}
  fmeh  := TFormEventHandlerClass.Create;
  try
    Form.KeyPreview := True;
    ClipboardText   := '';
    HasSelectorPnl  := False;
    HasProgressPnl  := False;
    HasButtonPnl    := False;
    HasFooterPnl    := False;
    pnlSelector     := nil;
    pnlExpand       := nil;
    if (InfoExpanded = '') then
      InfoExpanded   := 'Hide &Details';
    if (InfoCollapse = '') then
      InfoCollapse   := 'Show &Details';
    // initialize form properties
    Form.BorderStyle  := bsDialog;
    Form.BorderIcons  := [];
    if (Config.cButtons = 0) and (aCommonButtons = [cbOk]) then
      Include(aFlags, tdfAllowDialogCancellation); // just OK -> Esc/Alt+F4 close
    if (tdfAllowDialogCancellation in aFlags) then
      Form.BorderIcons := Form.BorderIcons + [biSystemMenu];
    if (tdfPositionRelativeToWindow in aFlags) then
      Form.Position := poOwnerFormCenter
    else
      Form.Position := poScreenCenter;
    Form.Font         := DefaultFont;
    FontHeight        := Form.Font.Height;
    PreCalcWidth;
    if (aWidth = 0) then
      aWidth := CalcDialogWidth;
    Form.ClientWidth  := aWidth;
    Form.ClientHeight := 120;
    Form.Caption      := Title;
    AddClipboardText('[Window Title]', Title, False);

    // create a white panel for the main dialog part
    Panel             := TPanel.Create(Form);
    Panel.Parent      := Form;
    Panel.Align       := alTop;
    Panel.BorderStyle := bsNone;
    Panel.BevelOuter  := bvNone;
    if ThemesEnabled then
      begin
        Panel.Color := clWindow;
        {$ifdef WITHUXTHEME}
        Panel.ParentBackground := False; // clWhite not used otherwise
        {$endif}
      end;
    Par := Panel;
    IconBorder := 10;
    if (tdfUseHIconMain in aFlags) or (WIN_ICONS[aDialogIcon] <> nil) then
      begin
        Image := TImage.Create(Form);
        Image.Parent := Par;
        if (tdfUseHIconMain in aFlags) then
          Image.Picture.Icon.Handle := MainIconHandle
        else
          Image.Picture.Icon.Handle := LoadIcon(0, WIN_ICONS[aDialogIcon]);
        Image.SetBounds(IconBorder, IconBorder, Image.Picture.Icon.Width, Image.Picture.Icon.Height);
        X := Image.Width + IconBorder * 2;
        Y := IconBorder;
      end
    else
      begin
        X := IconBorder;
        Y := IconBorder;
      end;
    if not ThemesEnabled then
      Inc(Y, 9);

    // add main texts (Instruction, Content, Information)
    AddLabel(Inst, True);
    AddClipboardText('[Main Instruction]', Inst, False);

    {$IF CompilerVersion >= 20.0}
    if (tdfEnableHyperLinks in aFlags) then
      AddLinkLabel(Content, False)
    else
    {$IFEND}
      AddLabel(Content, False);
    AddClipboardText('[Content]', Content, (tdfEnableHyperLinks in aFlags));
    if not ThemesEnabled then
      Dec(Y, 3);
    Panel.Height := Y;

    // add selection list or query editor
    if HAS_SELECTION then
      begin
        List := TStringList.Create;
        try
          Combo := TComboBox.Create(Form);
          Combo.Parent := Par;
          Combo.SetBounds(X + 6, Y, aWidth - 22 - X, 22);
          if (tdfQuery in aFlags) then
            Combo.Style := csDropDown
          else
            Combo.Style := csDropDownList;
          List.Text := Trim(Selection);
          Combo.Items.Assign(List);
          if HAS_QUERY then
            Combo.ItemIndex := List.IndexOf(Query)
          else
            Combo.ItemIndex := SelectionRes;
//          if (tdfQueryFieldFocused in aFlags) then
//            Form.ActiveControl := Combo;
          Panel.Height := Combo.Top + Combo.Height + 10;
        finally
          List.Free;
        end;
      end
    else if (tdfQuery in aFlags) then
      begin
        Edit := TEdit.Create(Form);
        Edit.Parent := Par;
        Edit.SetBounds(X + 6, Y, aWidth - 22 - X, 22);
        Edit.Text := Query;
        Edit.ImeMode := imClose;
        if (tdfQueryMasked in aFlags) then
          Edit.PasswordChar := '*';
//        if (tdfQueryFieldFocused in aFlags) then
//          Form.ActiveControl := Edit;
        Panel.Height := Edit.Top + Edit.Height + 14;
        PrevPanel := Panel;
        fmeh.Edit := Edit;
      end;
    PrevPanel := Panel;

    // add expand text (avobe)
    HasExpandPnl := HAS_INFO;
    IsExpandFooter := (tdfExpandFooterArea in aFlags);
    if HasExpandPnl then
      begin
        pnlExpand             := TPanel.Create(Form);
        pnlExpand.Parent      := Form;
        pnlExpand.Top         := GetPanelPosition;
        pnlExpand.Align       := alTop;
        pnlExpand.BorderStyle := bsNone;
        pnlExpand.BevelOuter  := bvNone;
        if ThemesEnabled then
          begin
            pnlExpand.Color := clWindow;
            {$ifdef WITHUXTHEME}
            pnlExpand.ParentBackground := False; // clWhite not used otherwise
            {$endif}
            Y := -2
          end
        else
          Y := 0;
        Par := pnlExpand;
        if IsExpandFooter then
          begin
            dX := X;
            if (Image <> nil) then
              Dec(X, Image.Width + IconBorder);
          end;
        dY := Y;
        {$IF CompilerVersion >= 20.0}
        if (tdfEnableHyperLinks in aFlags) then
          begin
            ExpandLinkLabel := AddLinkLabel(Info, False);
            ebeh.ExpandLinkLabel := ExpandLinkLabel;
          end
        else
        {$IFEND}
          begin
            ExpandLabel := AddLabel(Info, False);
            ebeh.ExpandLabel := ExpandLabel;
          end;
        if IsExpandFooter then
          X := dX;
        if not IsExpandFooter then
          AddClipboardText('[Expanded Information]', Info, (tdfEnableHyperLinks in aFlags));
        if ThemesEnabled then
          OffsetY :=  5
        else
          OffsetY := -5;
        ebeh.ExpandHeight := (Y - dY) + OffsetY;
        ebeh.CollapseHeight := 10;
        if (tdfExpandByDefault in aFlags) then
          pnlExpand.Height := ebeh.ExpandHeight
        else
          pnlExpand.Height := ebeh.CollapseHeight;
        ebeh.TopPos := GetPanelPosition - 4;
        PrevPanel := pnlExpand;
      end;

    //add progressbar
    if ShowProgress then
      begin
        HasProgressPnl          := True;
        pnlProgress             := TPanel.Create(Form);
        pnlProgress.Parent      := Form;
        pnlProgress.Top         := GetPanelPosition;
        pnlProgress.Align       := alTop;
        pnlProgress.BorderStyle := bsNone;
        pnlProgress.BevelOuter  := bvNone;
        if ThemesEnabled then
          begin
            pnlProgress.Color := clWindow;
            {$ifdef WITHUXTHEME}
            pnlProgress.ParentBackground := False; // clWhite not used otherwise
            {$endif}
          end;
        Par := pnlProgress;
        Progress                  := TProgressBarEx.Create(Form);
        Progress.Parent           := Par;
        FProgress.IsMarquee       := (tdfShowMarqueeProgressBar in aFlags);
        FProgress.ClientNonNative := Progress;
        if (tdfShowProgressBar in aFlags) then
          begin
            FProgress.Min            := Self.ProgressMin;
            FProgress.Max            := Self.ProgressMax;
            FProgress.Position       := Self.ProgressPosition;
            FProgress.State          := Self.ProgressState;

            Progress.Style           := pbstNormal;
            Progress.Min             := Self.ProgressMin;
            Progress.Max             := Self.ProgressMax;
            Progress.Position        := Self.ProgressPosition;
            Progress.State           := Self.ProgressState;
          end
        else
          begin
            FProgress.MarqueeSpeed   := Self.MarqueeSpeed;
            FProgress.State          := Self.ProgressState;

            Progress.Style           := pbstMarquee;
            Progress.MarqueeInterval := Self.MarqueeSpeed;
            Progress.State           := Self.ProgressState;
          end;
        Progress.SetBounds(X, 0, aWidth - X - 10, 15);
        pnlProgress.Height := 25;
        PrevPanel := pnlProgress;
      end;

    // RadioButton or CommandLink
    if HAS_COMMANDLINK and HAS_BUTTONS then
      HasSelectorPnl := True;
    if HAS_RADIOS then
      HasSelectorPnl := True;
    if HasSelectorPnl then
      begin
        pnlSelector             := TPanel.Create(Form);
        pnlSelector.Parent      := Form;
        pnlSelector.Top         := GetPanelPosition;
        pnlSelector.Align       := alTop;
        pnlSelector.BorderStyle := bsNone;
        pnlSelector.BevelOuter  := bvNone;
        if ThemesEnabled then
          begin
            pnlSelector.Color := clWindow;
            {$ifdef WITHUXTHEME}
            pnlSelector.ParentBackground := False; // clWhite not used otherwise
            {$endif}
          end;
        Par := pnlSelector;
        PrevPanel := pnlSelector;
      end;
    Y := 0;

    // add radio buttons
    if HAS_RADIOS then
      with TStringList.Create do
        try
          Text := SysUtils.Trim(Radios);
          SetLength(Rad, Count);
          if ThemesEnabled then
            OffsetY := 7
          else
            begin
//            OffsetY := 8;
              OffsetY := 14;
              Inc(Y, 4);
            end;
          for i := 0 to Count - 1 do
            begin
              Rad[i] := TRadioButton.Create(Form);
              with Rad[i] do
                begin
                  RAD[i].WordWrap := True;
                  Parent := Par;
                  Caption := StringReplace(Strings[i], '\n', #$0D#$0A, [rfReplaceAll]);
                  dmyDC := GetDC(RAD[i].Handle);
                  dmyCanvas := TCanvas.Create;
                  try
                    dmyCanvas.Handle := dmyDC;
                    dmyCanvas.Font.Assign(Rad[i].Font);
                    dmyRect := Rect(0, 0, aWidth - X - 12, 6 - FontHeight);
                    dmyRect.Bottom := DrawText(dmyCanvas.Handle, Pointer(Caption), -1, dmyRect, DT_WORD_ELLIPSIS or DT_CALCRECT or DT_WORDBREAK);
                  finally
                    ReleaseDC(RAD[i].Handle, dmyDC);
                    dmyCanvas.Free;
                  end;
                  SetBounds(X + 10, Y, aWidth - X - 12, dmyRect.Bottom);
                  Inc(Y, Height + OffsetY);
                  if (i = 0) or ((i + 200) = aRadioDef) then
                    Checked := True;
                end;
            end;
          if ThemesEnabled then
            Inc(Y, 10)
          else
            Inc(Y, 6);
          pnlSelector.Height := Y;
        finally
          Free;
        end;

    // add command links buttons
    if HAS_COMMANDLINK and HAS_BUTTONS then
      with TStringList.Create do
        try
          if not ThemesEnabled then
            Inc(Y, 3);
          Text := SysUtils.Trim(Buttons);
          for i := 0 to Count - 1 do
            begin
              CommandLink := TSynButton.Create(Form);
              with CommandLink do
                begin
                  Parent := Par;
                  if ThemesEnabled then
                    CommandLinkDefaultheight := 40
                  else
                    CommandLinkDefaultheight := 28;
                  SetBounds(X, Y, aWidth - X - 10, CommandLinkDefaultheight);
                  ShowGlyph := not (tdfUseCommandLinksNoIcon in aFlags);
                  if ShowGlyph then
                    Glyph := BitmapArrow;
                  CommandLink.Caption := NoCR(Strings[i]);
                  if (aHint <> '') then
                    CommandLinkHint := aHint; // note shown as Hint
                  AdjustHeight;
                  Inc(Y, CommandLink.Height);
                  ModalResult := i + 100;
                  if (ModalResult = aButtonDef) then
                    begin
                      CommandLink.Default := True;
                      Form.ActiveControl := CommandLink;
                    end;
                end;
            end;
          if ThemesEnabled then
            Inc(Y, 10)
          else
            Inc(Y, 6);
          pnlSelector.Height := Y;
        finally
          Free;
        end;

    // add buttons and verification checkbox
    if ((Byte(aCommonButtons) <> 0) or HAS_VERIFY or (HAS_BUTTONS and (not HAS_COMMANDLINK))) or HasExpandPnl then
      begin
        CTLine                 := '';
        HasButtonPnl           := True;
        pnlButtons             := TPanel.Create(Form);
        pnlButtons.Parent      := Form;
        pnlButtons.Top         := GetPanelPosition;
        pnlButtons.Height      := 44;
        pnlButtons.Align       := alTop;
        pnlButtons.BorderStyle := bsNone;
        pnlButtons.BevelOuter  := bvNone;
//      pnlButtons.ParentBackGround := False; // debug
        Par := pnlButtons;
        AddLine(pnlButtons, True);
        CurrTabOrder := Panel.TabOrder;
        XB := aWidth;
        if not HAS_COMMANDLINK then
          with TStringList.Create do
            try
              Text := SysUtils.Trim(Buttons);
              for i := Count - 1 downto 0 do
                AddButton(Strings[i], i + 100);
            finally
              Free;
            end;
        if HAS_COMMANDLINK or (not HAS_BUTTONS) then
          for B := High(B) downto Low(B) do
            if (B in aCommonButtons) then
              AddButton(LoadResString(TD_BTNS(B)), TD_BTNMOD[B]);

        OExpandLength := ExpandLength;
        OVerifyLength := VerifyLength;

        ButtonPanelHeight := pnlButtons.Height;
        HeightSum := 0;
        if HAS_INFO then
          Inc(HeightSum, Math.Max(ExpandLength.Height +  3, 21));
        if HAS_VERIFY then
          Inc(HeightSum, Math.Max(VerifyLength.Height +  3, 21));
        if HeightSum > 26 then
          AdjustY := 2
        else
          AdjustY := 0;
        if HAS_INFO then
          begin
            // no information collapse/expand yet: it's always expanded
            Expand := TExpandButton.Create(Form);
            with Expand do
              begin
                Parent := pnlButtons;
                Expand := (tdfExpandByDefault in aFlags);
                CaptionExpand   := InfoExpanded;
                CaptionCollapse := InfoCollapse;
                OffsetY := -AdjustY;
                if not ThemesEnabled then
                  Inc(OffsetY, 3);
                OnClick := ebeh.Click;
              end;
            ExpandLength.Width  := ExpandLength.Width  + 34;
            ExpandLength.Height := Math.Max(ExpandLength.Height +  3, 21);
            Expand.SetBounds(10, OffsetY + 11, ExpandLength.Width, ExpandLength.Height);
            ebeh.ExpandButton := Expand;
            ebeh.ExpandPanel := pnlExpand;
            CtLine := CtLine + '[@ExpandButton@] @ExpandCaption@';
            if not HAS_VERIFY then
              pnlButtons.Height := Math.Max(ButtonPanelHeight, Expand.Top + OExpandLength.Height + 15 - (AdjustY * 2));
          end;

        if HAS_VERIFY then
          begin
            Verif := TExpandButton.Create(Form);
            with Verif do
              begin
                Parent := pnlButtons;
                Style := ebsCheckbox;
                if HAS_INFO then
                  OffsetY := ExpandLength.Height + 1
                else
                  OffsetY := 2;
              end;
            VerifyLength.Width  := VerifyLength.Width  + 34;
            VerifyLength.Height := Math.Max(VerifyLength.Height +  3, 21);
            pnlButtons.Height := pnlButtons.Height + (Math.Max(VerifyLength.Height, 24) - 24);
            with Verif do
              begin
                SetBounds(10, OffsetY + 11, VerifyLength.Width, VerifyLength.Height);
                CaptionCheck := Verify;
                Checked := VerifyChecked;
                SendToBack;
                CtLine := CtLine + '  [@VerifyChecked@] ' + Verify;
              end;
            if (not HAS_INFO) and (AdjustY = 0) then
              pnlButtons.Height := ButtonPanelHeight
            else
              pnlButtons.Height := Math.Max(ButtonPanelHeight, Verif.Top + OVerifyLength.Height + 15 - (AdjustY * 2));
          end;
        Inc(Y, pnlButtons.Height);
        PrevPanel := pnlButtons;

        CtLine := CtLine + ' ';
        with TStringList.Create do
          try
            Text := SysUtils.Trim(Buttons);
            for i := 0 to Count - 1 do
              ctLine := ctLine + ' [' + NoCR(Strings[i]) + ']';
          finally
            Free;
          end;
        for B := Low(B) to High(B) do
          if (B in aCommonButtons) then
            ctLine := ctLine + ' [' + LoadResString(TD_BTNS(B)) + ']';
        ClipboardText := ClipboardText + Trim(ctLine) + #$0D#$0A#$0D#$0A;
      end
    else
      XB := 0;

    // add footer text with optional icon
    if HAS_FOOTER then
      begin
        if ThemesEnabled then
          OffsetY := 0
        else
          OffsetY := -1;
        HasFooterPnl          := True;
        pnlFooter             := TPanel.Create(Form);
        pnlFooter.Parent      := Form;
        pnlFooter.Top         := GetPanelPosition;
        pnlFooter.Align       := alTop;
        pnlFooter.BorderStyle := bsNone;
        pnlFooter.BevelOuter  := bvNone;
//      pnlFooter.ParentBackGround := False; // debug
        AddLine(pnlFooter, False);
        Par := pnlFooter;
        if (tdfUseHIconFooter in aFlags) or (WIN_FOOTERICONS[aFooterIcon] <> nil) then
          begin
            SmallIconSize := GetSystemMetrics(SM_CXSMICON);
            Image        := TImage.Create(Form);
            Image.Parent := Par;
            Pic := TIcon.Create;
            Bmp := TBitmap.Create;
            try
              if (tdfUseHIconFooter in aFlags) then
                Pic.Handle := FooterIconHandle
              else
//              Pic.Handle := LoadIcon(0, WIN_FOOTERICONS[aFooterIcon]);
              Pic.Handle := LoadImage(GetModuleHandle(user32), MAKEINTRESOURCE(WIN_FOOTERICONS_SMALL[aFooterIcon]), IMAGE_ICON, SmallIconSize, SmallIconSize, LR_DEFAULTCOLOR);
              Bmp.Transparent        := True;
              Bmp.Canvas.Brush.Color := pnlFooter.Color;
              Bmp.Width              := SmallIconSize;
              Bmp.Height             := SmallIconSize;
              DrawIconEx(Bmp.Canvas.Handle, 0, 0, Pic.Handle, Bmp.Width, Bmp.Height, 0,
                Bmp.Canvas.Brush.Handle, DI_NORMAL);
              Image.Picture.Bitmap   := Bmp;
              Image.SetBounds(12, OffsetY + 11, Bmp.Width, Bmp.Height);
              X := Bmp.Width + 19;
            finally
              Bmp.Free;
              Pic.Free;
            end;
          end
        else
          X := 10;
        Y := OffsetY + 11;
        dY := Y;
        {$IF CompilerVersion >= 20.0}
        if (tdfEnableHyperLinks in aFlags) then
          AddLinkLabel(Footer, False)
        else
        {$IFEND}
          AddLabel(Footer, False);
        AddClipboardText('[Footer]', Footer, (tdfEnableHyperLinks in aFlags));
        if ThemesEnabled then
          OffsetY := 5
        else
          OffsetY := 8;
        pnlFooter.Height := (Y - dY) + OffsetY;
        PrevPanel := pnlFooter;
      end;

    // add expand text (below)
    if HasExpandPnl and IsExpandFooter then
      begin
        if HAS_INFO then
          AddClipboardText('[Expanded Information]', Info, (tdfEnableHyperLinks in aFlags));
        AddLine(pnlExpand, False);
        pnlExpand.Align := alNone;
        pnlExpand.Top   := GetPanelPosition;
        pnlExpand.Align := alTop;
        pnlExpand.ParentBackground := True;

        {$IF CompilerVersion >= 20.0}
        if (tdfEnableHyperLinks in aFlags) then
          begin
            ExpandLinkLabel.Top := 11;
            ExpandLinkLabel.ParentColor := False;
            ExpandLinkLabel.Color := Form.Color;
            ebeh.ExpandHeight   := ExpandLinkLabel.Top + ExpandLinkLabel.Height + 9;
          end
        else
        {$IFEND}
          begin
            ExpandLabel.Top := 11;
            ExpandLabel.Transparent := True;
            ebeh.ExpandHeight   := ExpandLabel.Top + ExpandLabel.Height + 9;
          end;
        ebeh.CollapseHeight := 0;
        if (tdfExpandByDefault in aFlags) then
          pnlExpand.Height := ebeh.ExpandHeight
        else
          pnlExpand.Height := ebeh.CollapseHeight;
        ebeh.TopPos := GetPanelPosition - 4;
        PrevPanel := pnlExpand;
      end;

    if HasExpandPnl then
      begin
        {$IF CompilerVersion >= 20.0}
        if (tdfEnableHyperLinks in aFlags) then
          ExpandLinkLabel.Visible := (tdfExpandByDefault in aFlags)
        else
        {$IFEND}
          ExpandLabel.Visible := (tdfExpandByDefault in aFlags);
      end;

    if (tdfQueryFieldFocused in aFlags) then
      begin
        if HAS_SELECTION then
          Form.ActiveControl := Combo
        else if (tdfQuery in aFlags) then
          Form.ActiveControl := Edit;
      end;

    // display the form
    Form.ClientHeight := GetPanelPosition;
    SetPositionToCenter;

    // retrieve the results
    AltDialog := Form;

    fmeh.Text         := ClipboardText;
    fmeh.ExpandButton := Expand;
    fmeh.CheckBox     := Verif;
    Form.OnKeyPress   := fmeh.FormKeyPress;

    result := Form.ShowModal;
    if (Combo <> nil) then
      begin
        SelectionRes := Combo.ItemIndex;
        Query        := Combo.Text;
      end
    else if (Edit <> nil) then
      Query := Edit.Text;
    if (Verif <> nil) then
      VerifyChecked := Verif.Checked;
    RadioRes := 0;
    for i := 0 to High(Rad) do
      if Rad[i].Checked then
        RadioRes := i + 200;
  finally
    fmeh.Free;
    {$IF CompilerVersion >= 20.0}
    lleh.Free;
    {$IFEND}
    ebeh.Free;
    if ShowProgress then
      FProgress.Free;
    Form.Free;
  end;
end;

procedure TTaskDialog.InitProgress;
begin
  FProgress      := nil;
  FMin           :=   0;
  FMax           := 100;
  FPosition      :=   0;
  FMarqueeSpeed  :=  10;
  FState         := pbsNormal;
end;

procedure TTaskDialog.SetMarqueeSpeed(const Value: Cardinal);
begin
  FMarqueeSpeed := Value;
  if (FProgress <> nil) then
    FProgress.MarqueeSpeed := FMarqueeSpeed;
end;

procedure TTaskDialog.SetMax(const Value: Integer);
begin
  FMax := Value;
  if (FProgress <> nil) then
    FProgress.Max := FMax;
end;

procedure TTaskDialog.SetMin(const Value: Integer);
begin
  FMin := Value;
  if (FProgress <> nil) then
    FProgress.Min := FMin;
end;

procedure TTaskDialog.SetPosition(const Value: Integer);
begin
  FPosition := Value;
  if (FProgress <> nil) then
    FProgress.Position := FPosition;
end;

procedure TTaskDialog.SetState(const Value: TProgressBarState);
begin
  FState := Value;
  if (FProgress <> nil) then
    FProgress.State := FState;
end;

{ TTaskDialogEx }

function TTaskDialogEx.Execute(aParent: HWND): Integer;
begin
  Result := Base.Execute(CommonButtons, ButtonDef, Flags, DialogIcon, FooterIcon,
    RadioDef, Width, aParent, NonNative, EmulateClassicStyle);
end;

procedure TTaskDialogEx.Init;
begin
  CommonButtons       := [];
  ButtonDef           := 0;
  Flags               := [];
  DialogIcon          := tiInformation;
  FooterIcon          := tfiWarning;
  RadioDef            := 0;
  Width               := 0;
  NonNative           := False;
  EmulateClassicStyle := False;
  Base.InitProgress;
end;

{$endif USETMSPACK}

initialization
  DefaultFont := TFont.Create;
  DefaultFont.Assign(Screen.MenuFont);
  {$ifndef USETMSPACK}
  InitComCtl6;
  Assert(Ord(tdfCanBeMinimized) = 15);
  {$endif USETMSPACK}
  BitmapOK := TBitmap.Create;
  BitmapOK.LoadFromResourceName(HInstance, 'btnOk'); // SQLite3btnok.bmp
  BitmapOK.Transparent    := True;
  BitmapArrow             := TBitmap.Create;
  BitmapArrow.LoadFromResourceName(HInstance, 'btnArrow'); // SQLite3btnArrow.bmp
  BitmapArrow.Transparent := True;
  {$IFNDEF UNICODE}
  @GetThemeFont := GetProcAddress(GetModuleHandle('uxtheme.dll'), 'GetThemeFont');
  {$ENDIF}

finalization
  DefaultFont.Free;
  BitmapArrow.Free;
  BitmapOK.Free;

end.
