unit uCommandLink;

interface

uses
  Classes, Windows, Messages, UxTheme, SysUtils, Math, Types,
  {$IF CompilerVersion < 20.0}
  DwmApi,
  {$IFEND}
  {$IF CompilerVersion >= 23.0}
  Vcl.Graphics, Vcl.Controls, Vcl.StdCtrls, Vcl.Themes, Vcl.Forms, Vcl.GraphUtil,
  {$ELSE}
  Graphics, Controls, StdCtrls, Themes, Forms, GraphUtil,
  {$IFEND}
  uWindowsVersion;

type
  {$IF CompilerVersion >= 20.0}
  TCommandLink = class(TCustomButton)
  {$ELSE}
  TCommandLink = class(TButton)
  {$IFEND}
  private
    FCanvas: TCanvas;
    FSpacing: Integer;
    FMargin: Integer;
    FMouseInControl: Boolean;
    FCommandLinkHint: string;
    FShowGlyph: Boolean;
    FGlyphSize: TSize;
    FGlyph: TBitmap;
    FCaption: string;
    procedure DrawItem(const DrawItemStruct: TDrawItemStruct);
    procedure SetMargin(Value: Integer);
    procedure CNMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure SetShowGlyph(const Value: Boolean);
    procedure SetGlyph(const Value: TBitmap);
  protected
    procedure CreateHandle; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure SetButtonStyle(ADefault: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AdjustHeight;
  published
    property Action;
    property Align;
    property Anchors;
    property BiDiMode;
    property Caption: string read FCaption write FCaption;
    property CommandLinkHint: string read FCommandLinkHint write FCommandLinkHint;
    property Constraints;
    property DoubleBuffered;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Glyph: TBitmap read FGlyph write SetGlyph;
    property Margin: Integer read FMargin write SetMargin default -1;
    property ParentBiDiMode;
    {$IF CompilerVersion >= 20.0}
    property ParentDoubleBuffered;
    {$IFEND}
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property ShowGlyph: Boolean read FShowGlyph write SetShowGlyph;
    property TabOrder;
    property TabStop;
    property Visible;
    property WordWrap;
    property OnClick;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseActivate;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

const
  {$EXTERNALSYM ODS_NOACCEL}
  ODS_NOACCEL = $100;
  {$EXTERNALSYM ODS_NOFOCUSRECT}
  ODS_NOFOCUSRECT = $200;

{$IFNDEF UNICODE}
var
  GetThemeFont: function(hTheme: HTHEME; hdc: HDC; iPartId, iStateId, iPropId: Integer;
    var pFont: LOGFONTW): HRESULT; stdcall;
{$ENDIF}

implementation

{ TCommandLink }

procedure TCommandLink.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TCommandLink.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TCommandLink.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if {ThemesEnabled and} not FMouseInControl and not (csDesigning in ComponentState) then
  begin
    FMouseInControl := True;
    Repaint;
  end;
end;

procedure TCommandLink.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if {ThemesEnabled and} FMouseInControl then
  begin
    FMouseInControl := False;
    Repaint;
  end;
end;

procedure TCommandLink.CNDrawItem(var Message: TWMDrawItem);
begin
  DrawItem(Message.DrawItemStruct^);
end;

procedure TCommandLink.CNMeasureItem(var Message: TWMMeasureItem);
var
  Temp: PMeasureItemStruct;
begin
  Temp := Message.MeasureItemStruct;
  with Temp^ do
  begin
    itemWidth := Width;
    itemHeight := Height;
  end;
end;

constructor TCommandLink.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TCanvas.Create;
  FSpacing := 4;
  FMargin := -1;
  ControlStyle := ControlStyle + [csClickEvents, csReflector];
  {$IF CompilerVersion >= 22.0}
  ControlStyle := ControlStyle + [csPaintBlackOpaqueOnGlass];
  {$IFEND}
  DoubleBuffered := True;
  FShowGlyph := True;
  FGlyphSize.cx := 20;
  FGlyphSize.cy := 20;
  FGlyph := TBitmap.Create;
end;

procedure TCommandLink.CreateHandle;
begin
  inherited CreateHandle;
end;

procedure TCommandLink.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or BS_OWNERDRAW;
end;

destructor TCommandLink.Destroy;
begin
  FGlyph.Free;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TCommandLink.DrawItem(const DrawItemStruct: TDrawItemStruct);
var
  bRect, dRect, cRect, CaptionRect, BodyRect: TRect;
  IsDown, IsFocus{, IsHot}, IsNoFocusRect, IsNoAccel: Boolean;
  Details: TThemedElementDetails;
  Button: TThemedButton;
  iButton, iGlyph: Integer;
  Flags: UInt;
  Text: string;
  TextFlags: Cardinal;
  PaintOnGlass: Boolean;
  Theme: hTheme;
  uType: Cardinal;
  ColorRef: TColorRef;
  CAPTION_LOGFONT, BODY_LOGFONT: {$IFDEF UNICODE}TLOGFONT{$ELSE}TLOGFONTW{$ENDIF};
  GlyphSize: TSize;
  {OffsetX, }OffsetY: Integer;
  function GetGlyphWidth: Integer;
  begin
    if FShowGlyph then
      result := FGlyphSize.cx + 6
    else
      result := 0;
  end;
begin
  cRect := ClientRect;
  with DrawItemStruct do
    begin
      FCanvas.Handle := hDC;
      FCanvas.Font   := Self.Font;
      IsDown         := ((itemState and ODS_SELECTED) <> 0);
      IsFocus        := ((itemState and ODS_FOCUS) <> 0);
//    IsHot          := ((itemState and ODS_HOTLIGHT) <> 0);
      IsNoFocusRect  := ((itemState and ODS_NOFOCUSRECT) <> 0);
      IsNoAccel      := ((itemState and ODS_NOACCEL) <> 0);
    end;
  // draw button
  dRect := DrawItemStruct.rcItem;
  bRect := dRect;
  if ThemesEnabled then
    begin
//    OffsetX := 0;
      OffsetY := 0;
      Theme   := 0;
      Button  := tbButtonDontCare;
      iButton := 0;
      iGlyph  := 0;
      if VistaOrLater then
        begin
          Theme := OpenThemeData(Handle,'BUTTON');
          if not Enabled then
            begin
              iButton := CMDLS_DISABLED;
              iGlyph  := CMDLGS_DISABLED;
            end
          else if IsDown then
            begin
              iButton := CMDLS_PRESSED;
              iGlyph  := CMDLGS_PRESSED;
            end
          else if FMouseInControl then
            begin
              iButton := CMDLS_HOT;
              iGlyph  := CMDLGS_HOT;
            end
          else if IsFocus then
            begin
              iButton := CMDLS_DEFAULTED;
              iGlyph  := CMDLGS_DEFAULTED;
            end
          else
            begin
              iButton := CMDLS_NORMAL;
              iGlyph  := CMDLGS_NORMAL;
            end;
        end
      else
        begin
          if not Enabled then
            Button := tbPushButtonDisabled
          else if IsDown then
            Button := tbPushButtonPressed
          else if FMouseInControl then
            Button := tbPushButtonHot
          else if IsFocus then
            Button := tbButtonDontCare
          else
            Button := tbPushButtonNormal;
        end;
      Details := Theme_Services.GetElementDetails(Button);

      // Parent background.
      {$IF CompilerVersion >= 20.0}
      PaintOnGlass := (csGlassPaint in ControlState);
      if not PaintOnGlass then
        Theme_Services.DrawParentBackground(Handle, FCanvas.Handle, Details, True)
      {$ELSE}
      PaintOnGlass := ThemesEnabled and DwmCompositionEnabled and
        not (csDesigning in ComponentState);
      if not PaintOnGlass then
        ThemeServices.DrawParentBackground(Handle, FCanvas.Handle, @Details, True)
      {$IFEND}
      else
        {$IF CompilerVersion >= 22.0}
        FillRect(FCanvas.Handle, cRect, GetStockObject(BLACK_BRUSH));
        {$ELSE}
        FillRect(FCanvas.Handle, cRect, Self.Brush.Handle);
        {$IFEND}
      if VistaOrLater then
        begin
          // Button shape.
          DrawThemeBackground(Theme, FCanvas.Handle, BP_COMMANDLINK, iButton, bRect, nil);
          if FShowGlyph then
            begin
              GetThemePartSize(Theme, FCanvas.Handle, BP_COMMANDLINKGLYPH, iGlyph, nil, TS_TRUE, GlyphSize);
              FGlyphSize := GlyphSize;
              bRect := Rect(0, 0, GlyphSize.cx, GlyphSize.cy);
              OffsetRect(bRect, 7, 10);
              DrawThemeBackground(Theme, FCanvas.Handle, BP_COMMANDLINKGLYPH, iGlyph, bRect, nil);
            end;
          GetThemeColor(Theme, BP_COMMANDLINK, iButton, TMT_TEXTCOLOR, ColorRef);
          GetThemeFont(Theme, FCanvas.Handle, BP_COMMANDLINK, iButton, TMT_FONT    , CAPTION_LOGFONT);
          GetThemeFont(Theme, FCanvas.Handle, BP_COMMANDLINK, iButton, TMT_BODYFONT, BODY_LOGFONT   );
          CloseThemeData(Theme);
        end
      else
        begin
          // Button shape.
          case Button of
            tbPushButtonHot:
              begin
//              Details := ThemeServices.GetElementDetails(tbPushButtonNormal);
//              ThemeServices.DrawElement(FCanvas.Handle, Details, bRect);
                InflateRect(bRect, -1, -1);
                GradientFillCanvas(FCanvas, clWhite, $00F0F0F0, bRect, gdVertical);
                FCanvas.Brush.Style := bsClear;
                FCanvas.Pen.Color := clBtnShadow;
                FCanvas.RoundRect(bRect.Left, bRect.Top, bRect.Right, bRect.Bottom, 6, 6);
                FCanvas.Brush.Style := bsSolid;
              end;
            tbPushButtonPressed:
              begin
                Details := Theme_Services.GetElementDetails(tbPushButtonDisabled);
                Theme_Services.DrawElement(FCanvas.Handle, Details, bRect);
              end;
            tbButtonDontCare: // Focus
              begin
                InflateRect(bRect, -1, -1);
                FCanvas.Brush.Style := bsClear;
                FCanvas.Pen.Color := $A49500;
                FCanvas.RoundRect(bRect.Left, bRect.Top, bRect.Right, bRect.Bottom, 6, 6);
                FCanvas.Brush.Style := bsSolid;
              end;
          end;

          if FShowGlyph and Assigned(FGlyph) then
            FCanvas.Draw(9, 14, FGlyph);

          case Button of
            tbPushButtonDisabled:
              ColorRef := $009C758E;
            tbPushButtonPressed:
              ColorRef := $00732006;
            tbPushButtonHot:
              ColorRef := $00E54A07;
          else
            ColorRef := $00551C15;
          end;
        end;
    end
  else
    begin
//    OffsetX := 3;
      OffsetY := 3;
      if FMouseInControl or IsDown then
        begin
          Flags := DFCS_ADJUSTRECT;
          uType := DFC_BUTTON;
          Flags := Flags + DFCS_BUTTONPUSH;
          if IsDown then
            Flags := Flags or DFCS_PUSHED;
          // Parent background.
          FillRect(FCanvas.Handle, cRect, Parent.Brush.Handle);
          // Button shape.
          DrawFrameControl(FCanvas.Handle, bRect, uType, Flags);
        end
      else if IsFocus then
        begin
          FCanvas.Brush.Style := bsClear;
          FCanvas.Pen.Color := clWindowText;
          FCanvas.Rectangle(bRect);
          FCanvas.Brush.Style := bsSolid;
        end;
      if FShowGlyph and Assigned(FGlyph) then
        FCanvas.Draw(6, 6, FGlyph);
    end;

  // draw focus rect
  if (not IsNoFocusRect) and IsFocus then
    begin
      bRect := dRect;
      InflateRect(bRect, -3, -3);
      FCanvas.Pen.Color   := clWindowFrame;
//    FCanvas.Brush.Color := clBtnFace;
      DrawFocusRect(FCanvas.Handle, bRect);
    end;

  // draw text
  SetBkMode(FCanvas.Handle, Ord(TRANSPARENT));

  // caption text
  Text := Caption;
  TextFlags := DT_VCENTER or DT_NOCLIP or DT_WORDBREAK;
  if IsNoAccel then
    TextFlags := TextFlags or DT_HIDEPREFIX;
  if ThemesEnabled then
    begin
      if VistaOrLater then
        FCanvas.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(CAPTION_LOGFONT)
      else
        begin
          FCanvas.Font.Assign(Self.Font);
          FCanvas.Font.Size  := FCanvas.Font.Size + 1;
        end;
      FCanvas.Font.Color := ColorRef and (not clSystemColor);
    end
  else
    begin
      FCanvas.Font.Assign(Self.Font);
      FCanvas.Font.Style := [fsBold];
      FCanvas.Font.Size  := FCanvas.Font.Size;
    end;
  CaptionRect := dRect;
  CaptionRect.Left   := CaptionRect.Left + GetGlyphWidth() + 7;
  CaptionRect.Top    := CaptionRect.Top  + 10 - OffsetY;
  if not ThemesEnabled then
    begin
      CaptionRect.Left   := CaptionRect.Left - 4;
      CaptionRect.Top    := CaptionRect.Top  - 1;
    end;
  CaptionRect.Right  := CaptionRect.Right  - 4;
  CaptionRect.Bottom := CaptionRect.Bottom - 4;
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), CaptionRect, TextFlags or DT_CALCRECT);
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), CaptionRect, TextFlags);

  // body text
  Text := CommandLinkHint;
  TextFlags := DT_NOPREFIX or DT_VCENTER or DT_NOCLIP or DT_WORDBREAK;
  if ThemesEnabled then
    begin
      if VistaOrLater then
        FCanvas.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(BODY_LOGFONT)
      else
        begin
          FCanvas.Font.Assign(Self.Font);
          FCanvas.Font.Size  := FCanvas.Font.Size + 1;
        end;
      FCanvas.Font.Color := ColorRef and (not clSystemColor);
    end
  else
    begin
      FCanvas.Font.Assign(Self.Font);
      FCanvas.Font.Style := [];
    end;
  BodyRect := dRect;
  BodyRect.Left   := BodyRect.Left + GetGlyphWidth() + 7;
  BodyRect.Top    := CaptionRect.Bottom;
  if not ThemesEnabled then
    begin
      BodyRect.Left   := BodyRect.Left - 4;
      BodyRect.Top    := BodyRect.Top  + 2;
    end;
  BodyRect.Right  := BodyRect.Right  - 4;
  BodyRect.Bottom := dRect.Bottom - 14 - (OffsetY * 4);
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), BodyRect, TextFlags or DT_CALCRECT);
  if (BodyRect.Bottom - BodyRect.Top) = 0 then
    BodyRect.Bottom := BodyRect.Bottom + 1;
  Self.Height := Math.Max(Self.Height, BodyRect.Bottom - dRect.Top + 10 - (OffsetY * 4));
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), BodyRect, TextFlags);

  FCanvas.Handle := 0;
end;

procedure TCommandLink.AdjustHeight;
var
  Theme: hTheme;
  dCanvas: TCanvas;
  dRect, CaptionRect, BodyRect: TRect;
  CAPTION_LOGFONT, BODY_LOGFONT: {$IFDEF UNICODE}TLOGFONT{$ELSE}TLOGFONTW{$ENDIF};
  TextFlags: Integer;
  DC: hdc;
  {OffsetX, }OffsetY: Integer;
  function GetGlyphWidth: Integer;
  begin
    if FShowGlyph then
      result := 20 + 6
    else
      result := 0;
  end;
begin
  dRect := Rect(0, 0, Self.Width, Self.Height);
  dCanvas := TCanvas.Create;
  try
    DC := GetDC(Self.Handle);
    dCanvas.Handle := DC;

    if ThemesEnabled then
      begin
//      OffsetX := 0;
        OffsetY := 0;
        if VistaOrLater then
          begin
            Theme := OpenThemeData(Handle,'BUTTON');
            GetThemeFont(Theme, dCanvas.Handle, BP_COMMANDLINK, CMDLS_NORMAL, TMT_FONT    , CAPTION_LOGFONT);
            GetThemeFont(Theme, dCanvas.Handle, BP_COMMANDLINK, CMDLS_NORMAL, TMT_BODYFONT, BODY_LOGFONT   );
            CloseThemeData(Theme);
          end;
      end
    else
      begin
//      OffsetX := 3;
        OffsetY := 3;
      end;

    Text := Caption;
    TextFlags := DT_VCENTER or DT_NOCLIP or DT_WORDBREAK;
    if ((Perform(WM_QUERYUISTATE, 0, 0) and UISF_HIDEACCEL) <> 0) then
      TextFlags := TextFlags or DT_HIDEPREFIX;
    if ThemesEnabled then
      begin
        if VistaOrLater then
          dCanvas.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(CAPTION_LOGFONT)
        else
          begin
            dCanvas.Font.Assign(Self.Font);
            dCanvas.Font.Size  := dCanvas.Font.Size + 1;
          end;
      end
    else
      begin
        dCanvas.Font.Assign(Self.Font);
        dCanvas.Font.Style := [fsBold];
        dCanvas.Font.Size  := dCanvas.Font.Size + 1;
      end;

    CaptionRect := dRect;
    CaptionRect.Left   := CaptionRect.Left + GetGlyphWidth() + 7;
    CaptionRect.Top    := CaptionRect.Top  + 10 - OffsetY;
    if not ThemesEnabled then
      begin
        CaptionRect.Left   := CaptionRect.Left - 4;
        CaptionRect.Top    := CaptionRect.Top  + 2;
      end;
    CaptionRect.Right  := CaptionRect.Right  - 4;
    CaptionRect.Bottom := CaptionRect.Bottom - 4;
    Windows.DrawText(dCanvas.Handle, PChar(Text), Length(Text), CaptionRect, TextFlags or DT_CALCRECT);

    Text := CommandLinkHint;
    TextFlags := DT_NOPREFIX or DT_VCENTER or DT_NOCLIP or DT_WORDBREAK;
    if ThemesEnabled then
      begin
        if VistaOrLater then
          dCanvas.Font.Handle := {$IFDEF UNICODE}CreateFontIndirect{$ELSE}CreateFontIndirectW{$ENDIF}(BODY_LOGFONT)
        else
          begin
            dCanvas.Font.Assign(Self.Font);
            dCanvas.Font.Size  := dCanvas.Font.Size;
          end;
      end
    else
      begin
        dCanvas.Font.Assign(Self.Font);
        dCanvas.Font.Style := [];
      end;
    BodyRect := dRect;
    BodyRect.Left   := BodyRect.Left + GetGlyphWidth() + 7;
    BodyRect.Top    := CaptionRect.Bottom + 1;
    if not ThemesEnabled then
      begin
        BodyRect.Left   := BodyRect.Left - 4;
        BodyRect.Top    := BodyRect.Top  + 2;
      end;
    BodyRect.Right  := BodyRect.Right  - 4;
    BodyRect.Bottom := dRect.Bottom - 14 - (OffsetY * 4);
    Windows.DrawText(dCanvas.Handle, PChar(Text), Length(Text), BodyRect, TextFlags or DT_CALCRECT);
    if (BodyRect.Bottom - BodyRect.Top) = 0 then
      BodyRect.Bottom := BodyRect.Bottom + 1;
    Self.Height := Math.Max(Self.Height, BodyRect.Bottom - dRect.Top + 10 - (OffsetY * 4));
    ReleaseDC(Self.Handle, DC);
  finally
    dCanvas.Free;
  end;
end;

procedure TCommandLink.SetButtonStyle(ADefault: Boolean);
begin
  //  This is necessary for Owner-Draw
end;

procedure TCommandLink.SetGlyph(const Value: TBitmap);
begin
  FGlyph.Assign(Value);
  Invalidate;
end;

procedure TCommandLink.SetMargin(Value: Integer);
begin
  if (Value <> FMargin) and (Value >= - 1) then
  begin
    FMargin := Value;
    Invalidate;
  end;
end;

procedure TCommandLink.SetShowGlyph(const Value: Boolean);
begin
  FShowGlyph := Value;
  Invalidate;
end;

{$IFNDEF UNICODE}
initialization
  @GetThemeFont := GetProcAddress(GetModuleHandle('uxtheme.dll'), 'GetThemeFont');
{$ENDIF}

end.
