unit uExpandButton;

interface

uses
  Classes, Windows, Messages, UxTheme, SysUtils, Math, Types,
  {$IF CompilerVersion < 20.0}
  DwmApi,
  {$IFEND}
  {$IF CompilerVersion >= 23.0}
  Vcl.Graphics, Vcl.Controls, Vcl.StdCtrls, Vcl.Themes, Vcl.Forms,
  {$ELSE}
  Graphics, Controls, StdCtrls, Themes, Forms,
  {$IFEND}
  uWindowsVersion;

type
  TExpandButtonStyle = (ebsExpand, ebsCheckbox);

  {$IF CompilerVersion >= 20.0}
  TExpandButton = class(TCustomButton)
  {$ELSE}
  TExpandButton = class(TButton)
  {$IFEND}
  private
    FCanvas: TCanvas;
    FSpacing: Integer;
    FMargin: Integer;
    FMouseInControl: Boolean;
    FExpand: Boolean;
    FCaptionExpand: string;
    FCaptionCollapse: string;
    FStyle: TExpandButtonStyle;
    procedure SetCaption;
    procedure Toggle;
    procedure DrawItem(const DrawItemStruct: TDrawItemStruct);
    procedure SetMargin(Value: Integer);
    procedure CNMeasureItem(var Message: TWMMeasureItem); message CN_MEASUREITEM;
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure SetCaptionCollapse(const Value: string);
    procedure SetCaptionExpand(const Value: string);
    procedure SetExpand(const Value: Boolean);
  protected
    procedure CreateHandle; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure SetButtonStyle(ADefault: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Action;
    property Align;
    property Anchors;
    property BiDiMode;
    property CaptionCheck: string read FCaptionExpand write SetCaptionExpand;
    property Caption;
    property CaptionExpand: string read FCaptionExpand write SetCaptionExpand;
    property CaptionCollapse: string read FCaptionCollapse write SetCaptionCollapse;
//  property Checked;
    property Constraints;
    property DoubleBuffered;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Expand: Boolean read FExpand write SetExpand;
    property Checked: Boolean read FExpand write SetExpand;
    property Font;
    property Margin: Integer read FMargin write SetMargin default -1;
    property ParentBiDiMode;
    {$IF CompilerVersion >= 20.0}
    property ParentDoubleBuffered;
    {$IFEND}
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Style: TExpandButtonStyle read FStyle write FStyle;
    property TabOrder;
    property TabStop;
    property Visible;
    property WordWrap;
    property OnClick;
    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseActivate;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

const
  {$EXTERNALSYM ODS_NOACCEL}
  ODS_NOACCEL = $100;
  {$EXTERNALSYM ODS_NOFOCUSRECT}
  ODS_NOFOCUSRECT = $200;

implementation

{ TExpandButton }

procedure TExpandButton.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and CanFocus then
      begin
        case FStyle of
          ebsExpand:
            Toggle;
          ebsCheckbox:
            Perform(BM_CLICK, 0, 0);
        end;
        Result := 1;
        inherited;
      end
    else if (CharCode = $20) and Self.Focused and Enabled then
      begin
        Result := 1;
        Perform(BM_CLICK, 0, 0);
      end
    else
      inherited;
end;

procedure TExpandButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TExpandButton.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TExpandButton.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if ThemesEnabled and not FMouseInControl and not (csDesigning in ComponentState) then
  begin
    FMouseInControl := True;
    Repaint;
  end;
end;

procedure TExpandButton.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if ThemesEnabled and FMouseInControl then
  begin
    FMouseInControl := False;
    Repaint;
  end;
end;

procedure TExpandButton.CNCommand(var Message: TWMCommand);
begin
  if Message.NotifyCode = BN_CLICKED then
    Toggle;
end;

procedure TExpandButton.CNDrawItem(var Message: TWMDrawItem);
begin
  DrawItem(Message.DrawItemStruct^);
end;

procedure TExpandButton.CNMeasureItem(var Message: TWMMeasureItem);
var
  Temp: PMeasureItemStruct;
begin
  Temp := Message.MeasureItemStruct;
  with Temp^ do
  begin
    itemWidth := Width;
    itemHeight := Height;
  end;
end;

constructor TExpandButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TCanvas.Create;
  FStyle := ebsExpand;
  FSpacing := 4;
  FMargin := -1;
  FExpand := False;
  ControlStyle := ControlStyle + [csClickEvents, csReflector];
  {$IF CompilerVersion >= 22.0}
  ControlStyle := ControlStyle + [csPaintBlackOpaqueOnGlass];
  {$IFEND}
  DoubleBuffered := True;
//FHideAccel := True; // debug
end;

procedure TExpandButton.CreateHandle;
begin
  inherited CreateHandle;
end;

procedure TExpandButton.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or BS_OWNERDRAW;
end;

destructor TExpandButton.Destroy;
begin
  inherited Destroy;
  FCanvas.Free;
end;

procedure TExpandButton.DrawItem(const DrawItemStruct: TDrawItemStruct);
var
  bRect, dRect, cRect, dmyRect: TRect;
  IsDown, IsFocus{, IsHot}, IsNoFocusRect, IsNoAccel: Boolean;
  Details: TThemedElementDetails;
  Button: TThemedScrollBar;
  CheckButton: TThemedButton;
  iValue, iButton: Integer;
  Flags: UInt;
  Text: string;
  TextFlags: Cardinal;
  PaintOnGlass: Boolean;
  Theme: hTheme;
  uType: Cardinal;
begin
  cRect := ClientRect;
  with DrawItemStruct do
    begin
      FCanvas.Handle := hDC;
      FCanvas.Font   := Self.Font;
      IsDown         := ((itemState and ODS_SELECTED) <> 0);
      IsFocus        := ((itemState and ODS_FOCUS) <> 0);
//    IsHot          := ((itemState and ODS_HOTLIGHT) <> 0);
      IsNoFocusRect  := ((itemState and ODS_NOFOCUSRECT) <> 0);
      IsNoAccel      := ((itemState and ODS_NOACCEL) <> 0);
    end;

  // draw button
  dRect := DrawItemStruct.rcItem;
  bRect := dRect;
  bRect.Right  := bRect.Left + 19;
  bRect.Bottom := bRect.Top  + 21;
  if ThemesEnabled then
    begin
      Theme   := 0;
      Button  := tsScrollBarDontCare;
      iButton := 0;
      case FStyle of
        ebsExpand:
          begin
            if VistaOrLater then
              begin
                Theme := OpenThemeData(Handle,'TASKDIALOG');
                if FExpand then
                  begin
                    if not Enabled then
                      iButton := TDLGEBS_EXPANDEDNORMAL
                    else if IsDown then
                      iButton := TDLGEBS_EXPANDEDPRESSED
                    else if FMouseInControl then
                      iButton := TDLGEBS_EXPANDEDHOVER
                    else
                      iButton := TDLGEBS_EXPANDEDNORMAL;
                  end
                else
                  begin
                    if not Enabled then
                      iButton := TDLGEBS_NORMAL
                    else if IsDown then
                      iButton := TDLGEBS_PRESSED
                    else if FMouseInControl then
                      iButton := TDLGEBS_HOVER
                    else
                      iButton := TDLGEBS_NORMAL;
                  end;
                GetThemeInt(Theme, TDLG_EXPANDOBUTTON, iButton, TMT_WIDTH , iValue);
                bRect.Right  := bRect.Left+ iValue;
                GetThemeInt(Theme, TDLG_EXPANDOBUTTON, iButton, TMT_HEIGHT, iValue);
                bRect.Bottom := bRect.Top + iValue;
              end
            else
              begin
                if FExpand then
                  begin
                    if not Enabled then
                      Button := tsArrowBtnUpDisabled
                    else if IsDown then
                      Button := tsArrowBtnUpPressed
                    else if FMouseInControl then
                      Button := tsArrowBtnUpHot
                    else
                      Button := tsArrowBtnUpNormal;
                  end
                else
                  begin
                    if not Enabled then
                      Button := tsArrowBtnDownDisabled
                    else if IsDown then
                      Button := tsArrowBtnDownPressed
                    else if FMouseInControl then
                      Button := tsArrowBtnDownHot
                    else
                      Button := tsArrowBtnDownNormal;
                  end;
                bRect.Right  := bRect.Left + 19;
                bRect.Bottom := bRect.Top  + 21;
              end;
            Details := Theme_Services.GetElementDetails(Button);
          end;
        ebsCheckbox:
          begin
            if FExpand then // Checked
              begin
                if not Enabled then
                  CheckButton := tbCheckBoxCheckedDisabled
                else if IsDown then
                  CheckButton := tbCheckBoxCheckedPressed
                else if FMouseInControl then
                  CheckButton := tbCheckBoxCheckedHot
                else
                  CheckButton := tbCheckBoxCheckedNormal;
              end
            else
              begin         // UnChecked
                if not Enabled then
                  CheckButton := tbCheckBoxUncheckedDisabled
                else if IsDown then
                  CheckButton := tbCheckBoxUncheckedPressed
                else if FMouseInControl then
                  CheckButton := tbCheckBoxUncheckedHot
                else
                  CheckButton := tbCheckBoxUncheckedNormal;
              end;
            bRect.Right  := bRect.Left + 19;
            bRect.Bottom := bRect.Top  + 21;
            Details := Theme_Services.GetElementDetails(CheckButton);
          end;
      end;

      // Parent background.
      {$IF CompilerVersion >= 20.0}
      PaintOnGlass := (csGlassPaint in ControlState);
      if not PaintOnGlass then
        Theme_Services.DrawParentBackground(Handle, FCanvas.Handle, Details, True)
      {$ELSE}
      PaintOnGlass := ThemesEnabled and DwmCompositionEnabled and
        not (csDesigning in ComponentState);
      if not PaintOnGlass then
        ThemeServices.DrawParentBackground(Handle, FCanvas.Handle, @Details, True)
      {$IFEND}
      else
        {$IF CompilerVersion >= 22.0}
        FillRect(FCanvas.Handle, cRect, GetStockObject(BLACK_BRUSH));
        {$ELSE}
        FillRect(FCanvas.Handle, cRect, Self.Brush.Handle);
        {$IFEND}
      if VistaOrLater and (FStyle <> ebsCheckbox) then
        begin
          // Button shape.
          DrawThemeBackground(Theme, FCanvas.Handle, TDLG_EXPANDOBUTTON, iButton, bRect, nil);
          CloseThemeData(Theme);
        end
      else
        begin
          // Button shape.
          Theme_Services.DrawElement(FCanvas.Handle, Details, bRect);
        end;
    end
  else
    begin
      uType   := 0;
      Flags := DFCS_ADJUSTRECT;
      case FStyle of
        ebsExpand:
          begin
            uType := DFC_SCROLL;
            if FExpand then
              Flags := Flags + DFCS_SCROLLUP
            else
              Flags := Flags + DFCS_SCROLLDOWN;
          end;
        ebsCheckbox:
          begin
            uType := DFC_BUTTON;
            Flags := Flags + DFCS_BUTTONCHECK;
            if FExpand then
              Flags := Flags or DFCS_CHECKED;
          end;
      end;
      if IsDown then
        Flags := Flags or DFCS_PUSHED;
      if DrawItemStruct.itemState and ODS_DISABLED <> 0 then
        Flags := Flags or DFCS_INACTIVE;
      bRect.Right  := bRect.Left+ 19;
      bRect.Bottom := bRect.Top + 21;
      dmyRect := bRect;
      if (FStyle = ebsCheckbox) then
        begin
          OffsetRect(dmyRect, (19 - GetSystemMetrics(SM_CXMENUCHECK)) div 2,
                              (21 - GetSystemMetrics(SM_CYMENUCHECK)) div 2);
          dmyRect.Right  := dmyRect.Left + GetSystemMetrics(SM_CXMENUCHECK) + 1;
          dmyRect.Bottom := dmyRect.Left + GetSystemMetrics(SM_CYMENUCHECK) + 1;
        end;
      // Parent background.
      FillRect(FCanvas.Handle, cRect, Parent.Brush.Handle);
      // Button shape.
      DrawFrameControl(FCanvas.Handle, dmyRect, uType, Flags);
    end;

  if ThemesEnabled then
    begin
      InflateRect(bRect, 0, 1);
      bRect.Left  := bRect.Right + 4;
      bRect.Right := dRect.Right;
      InflateRect(bRect, -2, -2);
      bRect.Right := bRect.Right - 7;
    end
  else
    begin
      InflateRect(bRect, 0, 1);
      bRect.Left  := bRect.Right + 4;
      bRect.Right := dRect.Right;
      InflateRect(bRect, -4, 0);
      bRect.Right := bRect.Right - 5;
    end;

  Text := Caption;
  TextFlags := DT_VCENTER or DT_NOCLIP or DT_WORDBREAK;
  if IsNoAccel then
    TextFlags := TextFlags or DT_HIDEPREFIX;

  bRect.Top := bRect.Top - 1;

  if not ThemesEnabled then
    begin
      OffsetRect(bRect, -2, 2);
      if (FStyle = ebsCheckbox) then
        OffsetRect(bRect, 0, ((21 - GetSystemMetrics(SM_CYMENUCHECK)) div 2) - 2);
    end;

  // draw focus rect
  cRect := bRect;
  cRect.Right := FCanvas.TextWidth(Text) + 100;
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), cRect, TextFlags or DT_CALCRECT);
  bRect.Right  := cRect.Right + 4;
  if FStyle = ebsCheckbox then
    bRect.Top := bRect.Top + 2;
  bRect.Bottom := cRect.Bottom + 2;
  FCanvas.Pen.Color   := clWindowFrame;
  FCanvas.Brush.Color := clBtnFace;
  if (not IsNoFocusRect) and IsFocus then
    DrawFocusRect(FCanvas.Handle, bRect);
  if FStyle = ebsCheckbox then
    bRect.Top := bRect.Top - 2;

  // draw text
  OffsetRect(bRect, 2, 2);
  SetBkMode(FCanvas.Handle, Ord(TRANSPARENT));
  FCanvas.Font.Color := Self.Font.Color;
  Windows.DrawText(FCanvas.Handle, PChar(Text), Length(Text), bRect, TextFlags);

  FCanvas.Handle := 0;
end;

procedure TExpandButton.SetButtonStyle(ADefault: Boolean);
begin
  // This is necessary for Owner-Draw
end;

procedure TExpandButton.SetMargin(Value: Integer);
begin
  if (Value <> FMargin) and (Value >= - 1) then
  begin
    FMargin := Value;
    Invalidate;
  end;
end;

procedure TExpandButton.Toggle;
begin
  FExpand := not FExpand;
  SetCaption;
end;

procedure TExpandButton.SetCaption;
begin
  case FStyle of
    ebsExpand:
      begin
        if FExpand then
          Caption := CaptionExpand
        else
          Caption := CaptionCollapse;
      end;
    ebsCheckbox:
      Caption := CaptionCheck;
  end;
end;

procedure TExpandButton.SetCaptionCollapse(const Value: string);
begin
  FCaptionCollapse := Value;
  SetCaption;
end;

procedure TExpandButton.SetCaptionExpand(const Value: string);
begin
  FCaptionExpand := Value;
  SetCaption;
end;

procedure TExpandButton.SetExpand(const Value: Boolean);
begin
  FExpand := Value;
  SetCaption;
end;

end.
